const util = require('util');
import * as http from "http";
import * as fs from "fs";
import { Helper } from "codeceptjs";
const {I} = inject();

export class nonHeal extends Helper {

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then performs click action.
     * Throws error if any of the action fails in two attempts.
     * The second parameter is a context (CSS or XPath locator) to narrow the search.
     * ```js
     * // simple link
     * z.click('Logout');
     * // button of form
     * z.click('Submit');
     * // CSS button
     * z.click('#form input[type=submit]');
     * // XPath
     * z.click('//form/*[@type=submit]');
     * // link in context
     * z.click('Logout', '#nav');
     * // using strict locator
     * z.click({css: 'nav a.login'});
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string} context (optional, null by default) element to search in CSS|XPath|Strict locator.
     */
     static async click(locator: string, context?: string) 
    {
        
        try 
        {
            if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("("))
            {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);

               
                
                    await I.waitForClickable(locator);
                    if (util.isNullOrUndefined(context)) 
                    {
                        await I.click(locator);
                    }
                    else
                    {
                        await I.click(locator,context);
                    }
            }
            else
            {
                if (util.isNullOrUndefined(context)) 
                {
                        await I.click(locator);
                   
                }
                else
                {
                        await I.click(locator, context);
                    
                }
            }
        }
        catch (error) 
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("("))
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);

                  
                        
                        await I.waitForClickable(locator);
                        if (util.isNullOrUndefined(context)) 
                        {
                            await I.click(locator);
                        }
                        else
                        {
                            await I.click(locator,context);
                        }
                    
                }
                else
                {
                    if (util.isNullOrUndefined(context)) 
                    {
                            await I.click(locator);
                       
                    }
                    else
                    {
                            await I.click(locator, context);
                    }
                }
                            
                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("click : Error while clicking on element located by : " + locator);
            }
        }
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & enabled, then clears field value.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.clearField('Email');
     * z.clearField('user[email]');
     * z.clearField('#email');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     */
    static async clearField(locator: string) {
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                await I.waitForEnabled(locator);

                await I.clearField(locator);
            }
            else {
                await I.clearField(locator);
            }

        }
        catch (error) {
            try {

                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    await I.waitForEnabled(locator);


                    await I.clearField(locator);
                }
                else 
                {
                    await I.clearField(locator);
                }

                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("clearField : Error while clearing value in element located by : " + locator);

            }
        }
    }


    /**
     * Waits for element to be visible, Clears element value & fills given value.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * // by label
     * z.fillField('Email', 'hello@world.com');
     * // by name
     * z.fillField('password', secret('123456'));
     * // by CSS
     * z.fillField('form#login input[name=username]', 'John');
     * // or by strict locator
     * z.fillField({css: 'form#login input[name=username]'} 'John');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string} value
     */
    static async fillField(locator: string, value: any) {
        try {
            await this.clearField(locator);

            await I.fillField(locator, value);
        }
        catch (error) {
            try {

                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                await this.clearField(locator);

                await I.fillField(locator, value);
                await this.generateLog(timestamp, error);

            }
            catch (error) {

                throw new Error("fillField : Error while filling value in element located by : " + locator);
            }
        }
    }


    /**
     * Checks for the visibility of text, if optional context parameter given then searches for the visibility of text on the locator element.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.see('Welcome'); // text welcome on a page
     * z.see('Welcome', '.content'); // text inside .content div
     * z.see('Register', {css: 'form.register'}); // use strict locator
     * ```
     * @param {string} text
     * @param {string} context optional, element located by CSS|Xpath|strict locator in which to search for text.
     */
    static async see(text: any, context?: string) {
        try {
            if (util.isNullOrUndefined(context)) {
                await I.wait(2);
                await I.see(text);
            }
            else {
                await I.waitForVisible(context);
                await I.see(text, context);
            }
        }
        catch (error) 
        {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (util.isNullOrUndefined(context)) {
                    await I.wait(2);
                    await I.see(text);
                }
                else 
                {
                    await I.waitForVisible(context);
                    await I.see(text, context);
                }

                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("see : Error while checking visibility of : " + text);
            }
        }
    }

    /**
     * Checks the visibilty of an element.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.seeElement('#modal');
     * ```
     * @param {string} locator CSS|XPath|strict locator. 
     */
    static async seeElement(locator: string) {
        try {

            await I.waitForVisible(locator);

            await I.seeElement(locator);
        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                await I.waitForVisible(locator);

                await I.seeElement(locator);
                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("seeElement : Error while checking visibility of : " + locator);
            }
        }
    }

    /**
     * Opens a webpage in browser with given url.
     * ```js
     * z.amOnPage('/'); // opens main page of website
     * z.amOnPage('https://github.com'); // opens github
     * z.amOnPage('/login'); // opens a login page
     * ```
     * @param {string} url 
     */
    static async amOnPage(url: string) {
        await I.amOnPage(url);
    }

    /**
     * Asserts that two values are equal. If they are not,an assertion error is thrown. 
     * ```js
     * z.assertEqual(expected,actual)
     * ```
     * @param {any} expected 
     * @param {any} actual 
     */
     static async assertEqual(expected: any, actual: any, msg?: string) {
        if (util.isNullOrUndefined(msg))
            await I.assertEqual(expected, actual);
        else
            await I.assertEqual(expected, actual, msg);
    }

    /**
     * Returns all window handles. Useful for referencing a specific handle when calling I.switchToWindow(handle)
     * ```js
     * const windows = await I.grabAllWindowHandles();
     * ```
     * @returns {Array} windows
     */
    static async grabAllWindowHandles(): Promise<any> {
        let windows = await I.grabAllWindowHandles();
        return windows;
    }

    /**
     * Returns current window handle
     * ```js
     * const window = await z.grabCurrentWindowHandle();
     * ```
     * @returns {any} window
     */
    static async grabCurrentWindowHandle(): Promise<any> {
        let window: any = await I.grabCurrentWindowHandle();
        return window;
    }

    /**
     * Grabs the number of visible elements
     * ```js
     * let numOfElements = await z.grabNumberOfVisibleElements('p');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @returns {any} totalElements
     */
    static async grabNumberOfVisibleElements(locator: string): Promise<any> {
        let totalElements = await I.grabNumberOfVisibleElements(locator);
        return totalElements;
    }

    /**
     * Open new tab and switch to it.
     * ```js
     * z.openNewTab();
     * ```
     */
    static async openNewTab() {
        await I.openNewTab();
    }

    /**
     * Presses a key or a key in combination with modifier keys in the browser (on a focused element).
     * ```js
    * z.pressKey('Backspace');
    * ```
    *
    * To press a key in combination with modifier keys, pass the sequence as an array. All modifier keys (`'Alt'`, `'Control'`, `'Meta'`, `'Shift'`) will be released afterwards.
    *
    * ```js
    * z.pressKey(['Control', 'Z']);
    * ```
    *
    * For specifying operation modifier key based on operating system it is suggested to use `'CommandOrControl'`.
    * This will press `'Command'` (also known as `'Meta'`) on macOS machines and `'Control'` on non-macOS machines.
    *
    * ```js
    * z.pressKey(['CommandOrControl', 'Z']);
    * ```
    *
    * Some of the supported key names are:
    * - `'AltLeft'` or `'Alt'`
    * - `'AltRight'`
    * - `'ArrowDown'`
    * - `'ArrowLeft'`
    * - `'ArrowRight'`
    * - `'ArrowUp'`
    * - `'Backspace'`
    * - `'Clear'`
    * - `'ControlLeft'` or `'Control'`
    * - `'ControlRight'`
    * - `'Command'`
    * - `'CommandOrControl'`
    * - `'Delete'`
    * - `'End'`
    * - `'Enter'`
    * - `'Escape'`
    * - `'F1'` to `'F12'`
    * - `'Home'`
    * - `'Insert'`
    * - `'MetaLeft'` or `'Meta'`
    * - `'MetaRight'`
    * - `'Numpad0'` to `'Numpad9'`
    * - `'NumpadAdd'`
    * - `'NumpadDecimal'`
    * - `'NumpadDivide'`
    * - `'NumpadMultiply'`
    * - `'NumpadSubtract'`
    * - `'PageDown'`
    * - `'PageUp'`
    * - `'Pause'`
    * - `'Return'`
    * - `'ShiftLeft'` or `'Shift'`
    * - `'ShiftRight'`
    * - `'Space'`
    * - `'Tab'`
     * @param {string|Array} key
     */
    static async pressKey(key: string | string[]) {
        await I.pressKey(key);
    }

    /**
     * Reload the current page.
     * ```js
     * z.refreshPage();
     * ```
     */
    static async refreshPage() {
        await I.refreshPage();
    }

    /**
     * Saves a screenshot to ouput folder (set in codecept.json or codecept.conf.js). 
     * Filename is relative to output folder. Optionally resize the window to the full available page scrollHeight and scrollWidth
     * to capture the entire page by passing true in as the second argument.
     * ```js
     * z.saveScreenshot('debug.png');
     * z.saveScreenshot('debug.png', true) //resizes to available scrollHeight and scrollWidth before taking screenshot
     * ``` 
     * @param {string} fileName
     * @param {boolean} fullPage optional
     */
    static async saveScreenshot(fileName: string, fullPage?: boolean) {
        if (util.isNullOrUndefined(fullPage)) {
            await I.saveScreenshot(fileName);
        }
        else {
            await I.saveScreenshot(fileName, fullPage);
        }
    }

    /**
     * Switches frame or in case of null locator reverts to parent.
     * ```js
     * z.switchTo('iframe'); // switch to first iframe
     * z.switchTo(); // switch back to main page
     * ```
     * @param {string} locator optional, CSS|XPath|strict locator.
     */
    static async switchTo(locator?: string) {
        if (!util.isNullOrUndefined(locator)) {
            await I.switchTo(locator);
        }
        else {
            await I.switchTo();
        }
    }

    /**
     * Switch focus to a particular tab by its number. It waits tabs loading and then switch tab.
     * ```js
     * z.switchToNextTab();
     * z.switchToNextTab(2);
     * ```
     * @param {number} tabNum optional, number of tabs to switch forward, default: 1.
     * @param {number} timeout optional, time in seconds to wait.
     */
    static async switchToNextTab(tabNum?: number, timeout?: number) {
        await I.switchToNextTab(tabNum, timeout);

    }

    /**
     * Switch to the window with a specified handle.
     * ```js
     * const windows = await z.grabAllWindowHandles();
     * //  do something
     * await z.switchToWindow( windows[0] );
     *
     * const window = await z.grabCurrentWindowHandle();
     * //  do something
     * await z.switchToWindow( window );
     * ``` 
     * @param {string} window
     */
    static async switchToWindow(window: any) {
            await I.switchToWindow(window);
    }

    /**
     * Pauses execution for a number of seconds.
     * ```js
     * z.wait(2); // wait 2 secs
     * ```
     * @param {number} timeout
     */
    static async wait(timeout: any) {
        await I.wait(timeout);
    }

    /**
     * Waits for the visibility of element & for the specified value to be in value attribute.
     * ```js
     * z.waitForValue('//input', "GoodValue");
     * ```
     * @param {string} locator input field locator.
     * @param {string} value
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     */
    static async waitForValue(locator: string, value: string, timeout?: number) {
        try {
            await I.waitForVisible(locator);
            if (util.isNullOrUndefined(timeout)) {
                await I.waitForValue(locator, value);
            }
            else {
                await I.waitForValue(locator, value, timeout);

            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("waitForValue : Error while waiting for value in element located by : " + locator);
        }
    }

    /**
     * Waits for the visibility of element.
     * ```js
     * z.waitForVisible('#popup');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"  
     */
    static async waitForVisible(locator: string, timeout?: number) {
        try {
            if (!util.isNullOrUndefined(timeout)) {
                await I.waitForVisible(locator, timeout);
            }
            else {
                await I.waitForVisible(locator);
            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("waitForVisible : Error while waiting for visibility of nn element located by : " + locator);

        }
    }

    /**
     * Waits for a text to appear. 
     * Element can be located by CSS or XPath. Narrow down search results by providing context.
     * ```js
     * z.waitForText('Thank you, form has been submitted');
     * z.waitForText('Thank you, form has been submitted', 5, '#modal');
     * ```
     * @param {string} text
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     * @param {string} context optional, element located by CSS|XPath|strict locator.
     
     */
    static async waitForText(text: string, timeout?: number, context?: string) {
        try {
            if (!util.isNullOrUndefined(context) && util.isNullOrUndefined(timeout)) {
                await I.scrollIntoView(context,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForText(text, undefined, context);
            }
            else if (!util.isNullOrUndefined(timeout) && util.isNullOrUndefined(context)) {
                await I.waitForText(text, timeout);
            }
            else if (!util.isNullOrUndefined(timeout) && !util.isNullOrUndefined(context)) {
                await I.scrollIntoView(context, {block: "start"});
                await I.waitForText(text, timeout, context);
            }
            else if(util.isNullOrUndefined(context) && util.isNullOrUndefined(timeout))
            {
                await I.waitForText(text);
            }

        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");           
            throw new Error("waitForText : Error while waiting for text : " + text);

        }
    }

    /**
     * Waits for an element to be removed or become invisible on a page. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForInvisible('#popup');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     */
    static async waitForInvisible(locator: string, timeout?: number) {
        try {
            if (util.isNullOrUndefined(timeout)) {
                await I.waitForInvisible(locator);
            }
            else {
                await I.waitForInvisible(locator, timeout);
            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");
            throw new Error("waitForInvisible : Error while waiting invisibility of element located by: " + locator);

        }
    }

    /**
     * Waits for an element to become enabled. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForEnabled('#popup');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     */
    static async waitForEnabled(locator: string, timeout?: number) {
        try {
            if (util.isNullOrUndefined(timeout)) {
                await I.waitForEnabled(locator);
            }
            else {
                await I.waitForEnabled(locator, timeout);
            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");
            throw new Error("waitForEnabled : Error while waiting for element to be enabled located by: " + locator);

        }
    }

    /**
     * Waits for an element to be present on page. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForElement('.btn.continue');
     * z.waitForElement('.btn.continue', 5); // wait for 5 secs
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout" 
     */
    static async waitForElement(locator: string, timeout?: number) {
        try {
            if (!util.isNullOrUndefined(timeout)) {
                await I.waitForElement(locator, timeout);
            }
            else {
                await I.waitForElement(locator);

            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");
            throw new Error("waitForElement : Error while waiting for element to be present located by: " + locator);

        }
    }

    /**
     * Waits for an element to be clickable. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForClickable('.btn.continue');
     * z.waitForClickable('.btn.continue', 5); // wait for 5 secs
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"    
     */
    static async waitForClickable(locator: string, timeout?: number) {
        try {
            if (util.isNullOrUndefined(timeout)) {
                await I.waitForClickable(locator);
            }
            else {
                await I.waitForClickable(locator, timeout);
            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");

            throw new Error("waitForClickable : Error while waiting for element to be clickable located by: " + locator);

        }
    }

    /**
     * Waits for an element to be visible. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.attachFile('Avatar', 'data/avatar.jpg');
     * z.attachFile('form input[name=avatar]', 'data/avatar.jpg');
     * ```
     * @param {string} locator label|name|CSS|XPath|strict locator.
     * @param {string} filepath local file path relative to codecept.json config file
     */
    static async attachFile(locator: string, filepath: string) {
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.attachFile(locator, filepath);
            }
            else {
                await I.attachFile(locator, filepath);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.attachFile(locator, filepath);
                }
                else {
                    await I.attachFile(locator, filepath);
                }
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("attachFile : Error while uploading file to the element located by: " + locator);
            }
        }
    }

    /**
     * Switch focus to a particular tab by its number. It waits tabs loading and then switch tab.
     * ```js
     * z.switchToPreviousTab();
     * z.switchToPreviousTab(2);
     * ```
     * @param {number} tabNum optional, number of tabs to switch backward, default: 1.
     * @param {number} timeout optional, time in seconds to wait.
     */
    static async switchToPreviousTab(tabNum?: number, timeout?: number) {


        await I.switchToPreviousTab(tabNum, timeout);

    }

    /**
     * Close current tab.
     * ```js
     * z.closeCurrentTab();
     * ```
     */
    static async closeCurrentTab() {
        await I.closeCurrentTab();
    }

    /**
     * Close all tabs except for the current one.
     * ```js
     * z.closeOtherTabs();
     * ```
     */
    static async closeOtherTabs() {
        await I.closeOtherTabs();
    }

    /**
     * Scroll element into viewport.
     * ```js
     * z.scrollIntoView('#submit');
     * z.scrollIntoView('#submit', { behavior: "smooth", block: "center", inline: "center" });
     * z.scrollIntoView('#submit', true);
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {any} alignTop optional, or scrollIntoViewOptions (optional), 
     * see https://developer.mozilla.org/en-US/docs/Web/API/Element/scrollIntoView.
     
     */
    static async scrollIntoView(locator: string, alignTop?: any) {
        if (alignTop==null || alignTop==undefined) {
            await I.scrollIntoView(locator, alignTop);
        }
        else {
            await I.scrollIntoView(locator,{block: "start"})
        }
    }

    /**
     * Retrieves a text from an element located by CSS or XPath and returns it to test. 
     * Resumes test execution, so should be used inside static async with await operator.
     * If multiple elements found returns an array of texts.
     * ```js
     * let pin = await z.grabTextFrom('#pin');
     * ```
     * @param {string} locator element located by CSS|XPath|strict locator.
     * @returns {string|string[]} text 
     */
    static async grabTextFrom(locator: string): Promise<any> {
        let text: any = "";
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                text = await I.grabTextFrom(locator);
            }
            else {
                text = await I.grabTextFrom(locator);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    text = await I.grabTextFrom(locator);
                }
                else {
                    text = await I.grabTextFrom(locator);
                }
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("grabTextFrom : Error while fetching text from element located by: " + locator);
            }

        }

        return text;
    }

    static async grabTextFromAll(locator: string): Promise<any> {
        let text: any[] = [];
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                text = await I.grabTextFromAll(locator);
            }
            else {
                text = await I.grabTextFromAll(locator);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    text = await I.grabTextFromAll(locator);
                }
                else {
                    text = await I.grabTextFromAll(locator);
                }
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("grabTextFromAll : Error while fetching text from element located by: " + locator);
            }

        }

        return text;
    }

    /**
     * Retrieves an attribute from an element located by CSS or XPath and returns it to test. An array as a result will be returned if there are more than one matched element. 
     * Resumes test execution, so should be used inside static async with await operator.
     * ```js
     * let hint = await z.grabAttributeFrom('#tooltip', 'title');
     * ```
     * @param {string} locator element located by CSS|XPath|strict locator.
     * @param {string} attribute attribute name.
     * @returns {string} text 
     */
    static async grabAttributeFrom(locator: string, attribute: string, elementLocatorStrategy:string='xpath'): Promise<any> {
        let value: any = "";
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                
               
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
               
                await I.waitForVisible(locator);
                value = await this.grabPropertyOrAttributeValue(locator, attribute,elementLocatorStrategy);
    
            }
            else {
               
                value = await this.grabPropertyOrAttributeValue(locator, attribute,elementLocatorStrategy);
                
            }

        }
        catch (error) 
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                   
                        await I.waitForElement(locator);
                        await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                   
                    await I.waitForVisible(locator);
                    value = await this.grabPropertyOrAttributeValue(locator, attribute,elementLocatorStrategy);
                }
                else 
                {
                    value = await this.grabPropertyOrAttributeValue(locator, attribute,elementLocatorStrategy);
                }
                await this.generateLog(timestamp, error);
                
            }
            catch (error) 
            {
                throw new Error("grabAttributeFrom : Error while fetching attribute '" + attribute + "' from element located by: " + locator);
        }
        }

        return value;
    }

    static async grabAttributeFromAll(locator: string, attribute: string, elementLocatorStrategy: string = 'xpath'): Promise<any> {
        let value: any;
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                
             
                
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
              
                await I.waitForVisible(locator);
                value = await this.grabPropertyOrAttributeValueFromAll(locator, attribute, elementLocatorStrategy);
            }
            else {
                value = await this.grabPropertyOrAttributeValueFromAll(locator, attribute, elementLocatorStrategy);
            }

        }
        catch (error) 
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                  
                    
                        await I.waitForElement(locator);
                        await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                   
                        
                    await I.waitForVisible(locator);
                    value = await this.grabPropertyOrAttributeValueFromAll(locator, attribute, elementLocatorStrategy);
                }
                else 
                {
                    value = await this.grabPropertyOrAttributeValueFromAll(locator, attribute, elementLocatorStrategy);
                }
                await this.generateLog(timestamp, error);

            }
            catch (error) 
            {
                throw new Error("grabAttributeFromAll : Error while fetching attribute '" + attribute + "' from element located by: " + locator);
            }
        }

        return value;
    }


    /**
         * Scrolls element into viewport, waits for that element to become visible & clickable. Also checks if element is already checked.
         * Then selects a checkbox or radio button.
         * 
         * The second parameter is a context (CSS or XPath locator) to narrow the search.
         * ```js
         * z.checkOption('#agree');
         * z.checkOption('I Agree to Terms and Conditions');
         * z.checkOption('agree', '//form');
         * ```
         * @param {string} locator checkbox located by label | name | CSS | XPath | strict locator.
         * @param {string} context optional element located by CSS | XPath | strict locator.
         */
    static async checkOption(locator: string, context?: string) {
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                await I.waitForClickable(locator);
                await I.dontSeeCheckboxIsChecked(locator);
            }

            if (!util.isNullOrUndefined(context)) {
                await I.checkOption(locator, context);

            }
            else {
                await I.checkOption(locator);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    await I.waitForClickable(locator);
                    await I.dontSeeCheckboxIsChecked(locator);
                }

                if (!util.isNullOrUndefined(context)) {
                    await I.checkOption(locator, context);
                }
                else {
                    await I.checkOption(locator);
                }
                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("checkOption : Error while selecting element located by: " + locator);
            }
        }
    }

    /**
     * Checks that a text is not present on a page, if optional context parameter given then searches for the invisibility of text on the locator element.
     * Use context parameter to narrow down the search.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.dontSee('Login'); // assume we are already logged in.
     * z.dontSee('Login', '.nav'); // no login inside .nav element
     * ```
     * @param {string} text which is not present.
     * @param {string} context optional, element located by CSS|Xpath|strict locator in which to search for text.
     */
    static async dontSee(text: string, context?: string) {
        try {
            if (!util.isNullOrUndefined(context)) {
                await I.dontSee(text, context);
            }
            else {
                await I.dontSee(text);
            }
        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (!util.isNullOrUndefined(context)) {
                    await I.dontSee(text, context);
                }
                else {
                    await I.dontSee(text);
                }
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("dontSee : Error while checking invisibility of element by: " + text);
            }
        }
    }

    /**
     * Checks that element is not visible (or in DOM)
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.seeElement('#modal');
     * ```
     * @param {string} locator CSS|XPath|strict locator. 
     */
    static async dontSeeElement(locator: string) {
        try {            
                await I.dontSeeElement(locator);           
        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                await I.dontSeeElement(locator);
                
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("dontSeeElement : Error while checking invisibility of element by: " + locator);
            }
        }
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then performs double click action.
     * Throws error if any of the action fails in two attempts.
     * The second parameter is a context (CSS or XPath locator) to narrow the search.
     * ```js
     * z.doubleClick('Edit');
     * z.doubleClick('Edit', '.actions');
     * z.doubleClick({css: 'button.accept'});
     * z.doubleClick('.btn.edit');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string} context (optional, null by default) element to search in CSS|XPath|Strict locator.
     */
    static async doubleClick(locator: string) {
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                await I.waitForClickable(locator);
            }
           
                await I.doubleClick(locator);

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    await I.waitForClickable(locator);
                }
               
                    await I.doubleClick(locator);
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("doubleClick : Error while double clicking on element located by : " + locator);
            }
        }
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then Drag an item to a destination element.
     * Throws error if any of the action fails in two attempts.
     * 
     * ```js
     * z.dragAndDrop('#dragHandle', '#container');
     * ```
     * @param {string} srcLocator CSS|XPath|strict locator.
     * @param {string} destLocator CSS|XPath|Strict locator.
     */
    static async dragAndDrop(srcLocator: string, destLocator: string) {
        try {
            if (srcLocator.startsWith("/") || srcLocator.startsWith(".") || srcLocator.startsWith("#") || srcLocator.startsWith("{")) {
                await I.waitForElement(srcLocator);
                
                    await I.scrollIntoView(srcLocator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(srcLocator);
                    await I.waitForClickable(srcLocator);
                await I.waitForVisible(destLocator);
                await I.dragAndDrop(srcLocator, destLocator);
            }
            else {
                await I.dragAndDrop(srcLocator, destLocator);
            }

        }
        catch (error) 
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (srcLocator.startsWith("/") || srcLocator.startsWith(".") || srcLocator.startsWith("#") || srcLocator.startsWith("{")) 
                {
                    await I.waitForElement(srcLocator);
                    
                        await I.scrollIntoView(srcLocator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(srcLocator);
                   
                        await I.waitForClickable(srcLocator);
                    await I.waitForVisible(destLocator);
                    await I.dragAndDrop(srcLocator, destLocator);
                }
                else 
                {
                    await I.dragAndDrop(srcLocator, destLocator);
                }
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("dragAndDrop : Error while performing drag and drop from element located by : " + srcLocator + " to " + destLocator);
            }
        }
    }

    /**
     * Executes sync script on a page.
     * Pass arguments to function as additional parameters.
     * Will return execution result to a test.
     * In this case you should use static async function and await to receive results.
     *
     * Example with jQuery DatePicker:
     *
     * ```js
     * // change date of jQuery DatePicker
     * z.executeScript(function() {
     *   // now we are inside browser context
     *   $('date').datetimepicker('setDate', new Date());
     * });
     * ```
     * Can return values. Don't forget to use `await` to get them.
     *
     * ```js
     * let date = await z.executeScript(function(el) 
     *   // only basic types can be returned
     *   return $(el).datetimepicker('getDate').tostring();
     * } '#date'); // passing jquery selector
     * ```
     * @param {string|function} fn function to be executed in browser context.
     * @param {...any} args to be passed to function.     
     * */
    static async executeScript(fn: any | ((...params: any[]) => any), ...args: any[]): Promise<any> {
        let retunrtype = await I.executeScript(fn, args);
        return retunrtype;
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then selects an option in a drop-down select.
     * Throws error if any of the action fails in two attempts.
     * 
     * ```js
     * z.selectOption('Choose Plan', 'Monthly'); // select by label
     * z.selectOption('subscription', 'Monthly'); // match option by text
     * z.selectOption('subscription', '0'); // or by value
     * z.selectOption('//form/select[@name=account]','Premium');
     * z.selectOption('form select[name=account]', 'Premium');
     * z.selectOption({css: 'form select[name=account]'} 'Premium');
     * ```
     * Provide an array for the second argument to select multiple options.
     * ```js
     * z.selectOption('Which OS do you use?', ['Android', 'iOS']);
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string|string[]} option CSS|XPath|Strict locator.
     */
    static async selectOption(locator: string, option: string) {
        try {

            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                await I.waitForClickable(locator);
                await I.selectOption(locator, option);
            }
            else {
                await I.selectOption(locator, option);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    await I.waitForClickable(locator);
                    await I.selectOption(locator, option);
                }
                else {
                    await I.selectOption(locator, option);
                }
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("selectOption : Error while selecting value on element located by :  " + locator);
            }
        }
    }


    /**
     * Checks for element to be present, scrolls into view & Moves cursor to element matched by locator.
     * Extra shift can be set with offsetX and offsetY options.
     *
     * ```js
     * z.moveCursorTo('.tooltip');
     * z.moveCursorTo('#submit', 5,5);
     * ```
     *
     * @param {CodeceptJS.LocatorOrString} locator located by CSS|XPath|strict locator.
     * @param {number} [offsetX=0] (optional, `0` by default) X-axis offset.
     * @param {number} [offsetY=0] (optional, `0` by default) Y-axis offset.
     *
     *
     */
    static async moveCursorTo(locator: string,offsetX?:number,offsetY?:number) 
    {
        try 
        {

            if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
            {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});

            }
            
            if(util.isNullOrUndefined(offsetX) && util.isNullOrUndefined(offsetY))
            {
                await I.moveCursorTo(locator);
            }
            else if(util.isNullOrUndefined(offsetX) && !(util.isNullOrUndefined(offsetY)))
            {
                await I.moveCursorTo(locator,0,offsetY);
            }
            else if(!(util.isNullOrUndefined(offsetX)) && util.isNullOrUndefined(offsetY))
            {
                await I.moveCursorTo(locator,offsetX,0);
            }
            else
            {
                await I.moveCursorTo(locator,offsetX,offsetY);
            }
            

        }
        catch (error) 
        {
            try
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});

                }
                
                if(util.isNullOrUndefined(offsetX) && util.isNullOrUndefined(offsetY))
                {
                    await I.moveCursorTo(locator);
                }
                else if(util.isNullOrUndefined(offsetX) && !(util.isNullOrUndefined(offsetY)))
                {
                    await I.moveCursorTo(locator,0,offsetY);
                }
                else if(!(util.isNullOrUndefined(offsetX)) && util.isNullOrUndefined(offsetY))
                {
                    await I.moveCursorTo(locator,offsetX,0);
                }
                else
                {
                    await I.moveCursorTo(locator,offsetX,offsetY);
                }
                    
                await this.generateLog(timestamp, error);
            }
            catch (error) 
            {
                throw new Error("moveCursorTo : Error while moving cursor to element located by :  " + locator);
            }
        }
    }

    /**
     * Checks for element to be present, scrolls into view, waits for element to be visible and retrieves a value from a form element located by CSS or XPath, returns it to test.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let email = await z.grabValueFrom('input[name=email]');
     * ```
     * @param {CodeceptJS.LocatorOrString} locator field located by label|name|CSS|XPath|strict locator.
     * @returns {Promise<any>} attribute value
     *
     */
    static async grabValueFrom(locator: string, elementLocatorStrategy: string = 'xpath') :Promise<any>
    {
        let text: any = "";
        try 
        {

            if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
            {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                text = await this.grabPropertyOrAttributeValue(locator,'value', elementLocatorStrategy)
            }
            else 
            {
                text = await this.grabPropertyOrAttributeValue(locator,'value', elementLocatorStrategy)
            }

        }
        catch (error) 
        {
            try
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    text = await this.grabPropertyOrAttributeValue(locator,'value', elementLocatorStrategy)
                }
                else 
                {
                    text = await this.grabPropertyOrAttributeValue(locator,'value', elementLocatorStrategy)
                }
                
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("grabValueFrom : Error while grabbing value from element located by :  " + locator);
            }
        }
        return text;
    }

    static async grabValueFromAll(locator: string, elementLocatorStrategy: string = 'xpath') :Promise<any>
    {
        let text: any;
                try 
        {

            if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
            {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                text = await this.grabPropertyOrAttributeValueFromAll(locator,'value',elementLocatorStrategy)
            }
            else 
            {
                text = await this.grabPropertyOrAttributeValueFromAll(locator,'value',elementLocatorStrategy)
            }

        }
        catch (error) 
        {
            try
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if(locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    text = await this.grabPropertyOrAttributeValueFromAll(locator,'value', elementLocatorStrategy)
                }
                else 
                {
                    text = await this.grabPropertyOrAttributeValueFromAll(locator,'value', elementLocatorStrategy)
                }
                
                await this.generateLog(timestamp, error);
            }
            catch (error) {
                throw new Error("grabValueFromAll : Error while grabbing value from element located by :  " + locator);
            }
        }
        return text;
    }

    /**
     * Dismisses the active JavaScript popup, as created by window.alert|window.confirm|window.prompt.
     * ```js
     * await z.cancelPopup();
     * ```
     */
    static async cancelPopup() 
    {
        try 
        {
            await I.cancelPopup();  
        }
        catch (error) 
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("cancelPopup : Error while cancelling popup.");
        }
    }

    /**
     * Accepts the active JavaScript native popup window, as created by window.alert|window.confirm|window.prompt.
     * Don't confuse popups with modal windows, as created by [various
     * libraries](http://jster.net/category/windows-modals-popups).
     * ```js
     * await z.acceptPopup();
     * ```
     */
    static async acceptPopup() 
    {
        try 
        {
            await I.acceptPopup();  
        }
        catch (error) 
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("acceptPopup : Error while accepting popup.");
        }
    }

    /**
     * Checks that the active JavaScript popup, as created by `window.alert|window.confirm|window.prompt`, contains the
     * given string.
     *
     * @param {string} text value to check.
     * 
     * ```js
     * await z.seeInPopup("value");
     * ```
     * 
     */
    static async seeInPopup(text:string) 
    {
        try 
        {
            await I.seeInPopup(text);  
        }
        catch (error) 
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("seeInPopup : Error while checking text in popup.");
        }
    }

    /**
     * Gets a cookie object by name.
     * If none provided gets all cookies.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let cookie = await z.grabCookie('auth');
     * assert(cookie.value, '123456');
     * ```
     *
     * @param {?string} [name=null] cookie name.
     * @returns {Promise<string>} attribute value
     *
     */
    static async grabCookie(name:string): Promise<any>
    {
        let value:any = "";
        try 
        {
          value =  await I.grabCookie(name);  
        }
        catch (error) 
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("grabCookie : Error fetching cookie.");
        }
        return value
    }

    /**
     * Get current URL from browser
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let url = await z.grabCurrentUrl();
     * ```
     *
     */

    static async grabCurrentUrl(): Promise<any> {
        let url = "";

        try {
            url = await I.grabCurrentUrl();
        } catch (error) {
            let timestamp = new Date().getTime();
            await I.saveScreenshot(timestamp.toString() + ".png");
            await this.generateLog(timestamp, error);
            
            throw new Error("grabCurrentUrl : Error fetching current url.");
        }
        return url;
    }

     /**
     * Grab number of open tabs.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let tabs = await z.grabNumberOfOpenTabs();
     * ```
     *
     */

    static async grabNumberOfOpenTabs(): Promise<any> {
        try {
           
            let tabs = await I.grabNumberOfOpenTabs();
          
            return tabs;

        } catch (error) {
            let timestamp = new Date().getTime();
            await I.saveScreenshot(timestamp.toString() + ".png");
            await this.generateLog(timestamp, error);

            throw new Error("grabNumberOfOpenTabs : Error fetching number of open tabs.");
        }

    }

    /**
     * Checks that current url contains a provided fragment.
     *
     * @param {string} url value to check.
     * 
     * ```js
     * await z.seeInCurrentUrl("url");
     * ```
     * 
     */
    static async seeInCurrentUrl(url: string) {        
        try {
           
            await I.seeInCurrentUrl(url);

        } catch (error) {
            let timestamp = new Date().getTime();
            await I.saveScreenshot(timestamp.toString() + ".png");
            await this.generateLog(timestamp, error);

            throw new Error("seeInCurrentUrl : Error while seeing provided fragment "+ url+ " in current url");

        }

    }

    /**
     * Retrieves a page title and returns it to test.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let tabs = await z.grabTitle();
     * ```
     *
     */

    static async grabTitle(): Promise<any> {
        try {
            let title  = await I.grabTitle();            
            return title ;

        } catch (error) {
            let timestamp = new Date().getTime();
            await I.saveScreenshot(timestamp.toString() + ".png");
            await this.generateLog(timestamp, error);

            throw new Error("grabTitle : Error fetching title.");
        }

    }   

    /**
     * Scrolls element into viewport, waits for that element to become visible.
     * Then Verifies that the specified checkbox is not checked.
     * 
     * ```js
     * z.dontSeeCheckboxIsChecked('#agree');
     * z.dontSeeCheckboxIsChecked('I agree to terms');
     * z.dontSeeCheckboxIsChecked('agree');
     * ```
     * @param {string} field  located by label|name|CSS|XPath|strict locator.
     */
    static async dontSeeCheckboxIsChecked(field : string) {
        try {
            if (field.startsWith("/") || field.startsWith(".") || field.startsWith("#") || field.startsWith("{") || field.startsWith("(")) {
                await I.waitForElement(field);
                await I.scrollIntoView(field,{behavior: "auto", block: "center", inline: "center"});
                await I.dontSeeCheckboxIsChecked(field);
            }

            else {
                await I.dontSeeCheckboxIsChecked(field);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (field.startsWith("/") || field.startsWith(".") || field.startsWith("#") || field.startsWith("{") || field.startsWith("(")) {
                    await I.waitForElement(field);
                    await I.scrollIntoView(field,{behavior: "auto", block: "center", inline: "center"});
                    await I.dontSeeCheckboxIsChecked(field);
                }
    
                else {
                    await I.dontSeeCheckboxIsChecked(field);
                }

                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("dontSeeCheckboxIsChecked : Error while verfiying checkbox is not checked: " + field);
            }
        }
    }
    

        
    /**
     * Scrolls element into viewport, waits for that element to become visible.
     * Then Checks that the given input field or textarea equals to given value.
     * 
     * ```js
     * z.seeInField('Username', 'davert');
     * z.seeInField({css: 'form textarea'},'Type your comment here');
     * z.seeInField('#searchform input','Search');
     * ```
     * @param {string} field located by label|name|CSS|XPath|strict locator.
     * @param {string} value value to check.
     */
    static async seeInField(field : string, value: string) {
        try {
            if (field.startsWith("/") || field.startsWith(".") || field.startsWith("#") || field.startsWith("{") || field.startsWith("(")) {
                await I.waitForElement(field);
                await I.scrollIntoView(field,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(field);
                await I.seeInField(field, value);
            }

            else {
                await I.seeInField(field, value);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (field.startsWith("/") || field.startsWith(".") || field.startsWith("#") || field.startsWith("{") || field.startsWith("(")) {
                    await I.waitForElement(field);
                    await I.scrollIntoView(field,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(field);
                    await I.seeInField(field, value);
                }
    
                else {
                    await I.seeInField(field, value);
                }

                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("seeInField : Error while verfiying value in input field " + field);
            }
        }
    }

     /**
     * Asserts that two values are not equal. If they are ,an assertion error is thrown. 
     * ```js
     * z.assertNotEqual(expected,actual)
     * ```
     * @param {any} expected 
     * @param {any} actual 
     */
      static async assertNotEqual(expected: any, actual: any,msg?:string) {
        await I.assertNotEqual(expected, actual, msg);
    }
    
     /**
     * Scrolls to element matched by locator. Extra shift can be set with offsetX and offsetY options.
     * ```js
     * z.scrollTo('footer');
     * z.scrollTo('#submit', 5, 5);
     
     * ```
     * @param {string} locator located by CSS|XPath|strict locator.
     * @param {number} offsetX (optional, 0 by default) X-axis offset
     * @param {number} offsetY (optional, 0 by default) Y-axis offset.
     *
     
     */
    static async scrollTo(locator: string, offsetX?: number, offsetY?: number) {
       try 
       {
        if (!util.isNullOrUndefined(offsetX)) {
            await I.scrollTo(locator, offsetX);
        }
        else if (!util.isNullOrUndefined(offsetY)) {
            await I.scrollTo(locator, undefined ,offsetY);
        }
        else if (!util.isNullOrUndefined(offsetX) && !util.isNullOrUndefined(offsetY)) {
            await I.scrollTo(locator, offsetX, offsetY);
        }
        else {
            await I.scrollTo(locator);
        }

       } catch (error) 
       {
        let timestamp = new Date().getTime();
        await I.saveScreenshot(timestamp.toString() + ".png");
        await this.generateLog(timestamp, error);

        throw new Error("scrollTo : Error while scroll to element located by: " + locator);
       }
    }

    /**
     * Scroll page to the top.
     * ```js
     * z.scrollPageToTop();
     
     */
    static async scrollPageToTop() {
        try {
            await I.scrollPageToTop();
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await I.saveScreenshot(timestamp.toString() + ".png");
            await this.generateLog(timestamp, error);
            
            throw new Error("scrollPageToTop : Error while scroll page to top " );
        }
    }
    
    /**
     * Scroll page to the bottom.
     * ```js
     * z.scrollPageToBottom();
     
     */
    static async scrollPageToBottom() {
    try {
        await I.scrollPageToBottom();
    }
    catch (error) {
        let timestamp = new Date().getTime();
        await I.saveScreenshot(timestamp.toString() + ".png");
        await this.generateLog(timestamp, error);
        
        throw new Error("scrollPageToBottom : Error while scroll page to bottom " );
    }
    }

     /**
     * Asserts that an element is visible a given number of times
     * ```js
     * await z..seeNumberOfVisibleElements('.buttons', 3);
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} num number of elements..
     * 
     */
    static async seeNumberOfVisibleElements(locator: string, num: number){
      await I.seeNumberOfVisibleElements(locator, num);        
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible.
     * Then Verifies that the specified checkbox is checked.
     * 
     * ```js
     * z.seeCheckboxIsChecked('Agree');
     * z.seeCheckboxIsChecked('#agree'); 
     * z.seeCheckboxIsChecked({css: '#signup_form input[type=checkbox]'});
     * ```
     * @param {string} field  located by label|name|CSS|XPath|strict locator.
     */
    static async seeCheckboxIsChecked(field : string) {
        try {
            if (field.startsWith("/") || field.startsWith(".") || field.startsWith("#") || field.startsWith("{") || field.startsWith("(")) {
                await I.waitForElement(field);
                await I.scrollIntoView(field,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(field);
                await I.seeCheckboxIsChecked(field);
            }

            else {
                await I.seeCheckboxIsChecked(field);
            }

        }
        catch (error) {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (field.startsWith("/") || field.startsWith(".") || field.startsWith("#") || field.startsWith("{") || field.startsWith("(")) {
                    await I.waitForElement(field);
                    await I.scrollIntoView(field,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(field);
                    await I.seeCheckboxIsChecked(field);
                }
    
                else {
                    await I.seeCheckboxIsChecked(field);
                }

                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("seeCheckboxIsChecked : Error while verfiying checkbox is checked: " + field);
            }
        }
    }

     /**
     * Resize the current window to provided width and height. 
     * First parameter can be set to `maximize`.
     * ```js
     * z.resizeWindow(width, height)
     
     */
    static async resizeWindow(width: any, height?: any) {
    try {
        if(!util.isNullOrUndefined(height))
        {
            await I.resizeWindow(width,height);
        }
        else
        {
                await I.resizeWindow(width, 0);
        }
    }
    catch (error) {
        let timestamp = new Date().getTime();
        await I.saveScreenshot(timestamp.toString() + ".png");
        await this.generateLog(timestamp, error);
        
        throw new Error("resizeWindow : Error resizing the current window " );
        }
    }

      /**
       * Types out the given text into an active field. 
       * To slow down typing use a second parameter, to set interval between key presses. 
       * Note: Should be used when fillField is not an option.
       * ```js
       * z.type('Type this out.');
       * 
       * // typing values with a 100ms interval
       * z.type('4141555311111111', 100);
       * // When passing in an array
       * z.type(['T', 'E', 'X', 'T'])  
       *       
       * @param {string} key|keys  key or array of keys to type.
       * @param {number} delay (optional) delay in ms between key presses
       */
    static async type(key: string | string[], delay?: number) {
    try {
        if(util.isNullOrUndefined(delay))
        {
            await I.type(key)
        }
        else
        {
            await I.type(key, delay)
        }
        
    }
    catch (error) {
        let timestamp = new Date().getTime();
        await I.saveScreenshot(timestamp.toString() + ".png");
        await this.generateLog(timestamp, error);
        
        throw new Error("ERROR" + error + "\n" + key);
        }
    }

    /**
     * Performs a swipe Right inside an element. 
     * Can be slow or fast swipe.
     * ```js
     * z.swipeRight('#container');
     *       
     * @param {string} locator an element on which to perform swipe
     * @param {string}[speed='slow'] a speed to perform: `slow` or `fast`.
     */
    static async swipeRight(locator: string, speed?: string) {
    try {
        await I.swipeRight(locator, speed);   
    }
    catch (error) {
        let timestamp = new Date().getTime();
        await I.saveScreenshot(timestamp.toString() + ".png");
        await this.generateLog(timestamp, error);
        
        throw new Error("swipeRight : Error performing swipe right inside an element " );
        }
    }

    

    /**
     * Checks that title contains text.
     * ```js
     * z.seeInTitle('Home Page');
     * ```
     * @param {string} text value to check.
     */
    static async seeInTitle(text: any) {
        try {
            await I.seeInTitle(text);
        }

        catch (error) 
        {
            try {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("seeInTitle : Error while checking title contains text: " + text);
            }
        }
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & enabled, then Force clicks an element
     
     * ```js
     * z.forceClick('#hiddenButton');
     * z.forceClick('Click me', '#hidden');
     * ```
    * @param {string} locator CSS|XPath|strict locator.
     * @param {string} context (optional, null by default) element to search in CSS|XPath|Strict locator.
     */
    static async forceClick(locator: string, context?: string) {
        try {
            if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                await I.waitForElement(locator);
                await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                await I.waitForVisible(locator);
                await I.waitForEnabled(locator);
               
                if(util.isNullOrUndefined(context))
                {
                    await I.forceClick(locator);
                }
                else
                {
                    await I.forceClick(locator, context);
                }
            }
            else {
                if(util.isNullOrUndefined(context))
                {
                    await I.forceClick(locator);
                }
                else
                {
                    await I.forceClick(locator, context);
                }
            }   
        }
        catch (error) {
            try {

                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");

                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    await I.waitForEnabled(locator);
                   
                    if(util.isNullOrUndefined(context))
                    {
                        await I.forceClick(locator);
                    }
                    else
                    {
                        await I.forceClick(locator, context);
                    }
                }
                else {
                    if(util.isNullOrUndefined(context))
                    {
                        await I.forceClick(locator);
                    }
                    else
                    {
                        await I.forceClick(locator, context);
                    }
                } 

                await this.generateLog(timestamp, error);

            }
            catch (error) {
                throw new Error("forceClick : Error while clicking element located by : " + locator);

            }
        }
    }

    /**
     * Enters a directory In local filesystem.
     * Starts from a current directory
     * z.amInPath('output/downloads');
     * @param {string} openPath
     */
    static async amInPath(openPath: string)
    {
        try
        {
            await I.amInPath(openPath);
        }
        catch(error)
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");  

            throw new Error("amInPath: Error while entering a directory in filesysytem "+openPath);
        }    
    }
    /**
     * Waits for file to be present in current directory.
     *
     * ```js
     * z.handleDownloads();
     * z.click('Download large File');
     * z.amInPath('output/downloads');
     * z.waitForFile('largeFilesName.txt', 10); // wait 10 seconds for file
     * ```
     * @param {string} name
     * @param {number} [sec] seconds to wait
     */
    static async waitForFile(name: string, sec?: number)
    {
        try
        {
            if(!util.isNullOrUndefined(sec))
            {
                await I.waitForFile(name, sec);
            }
            else
            {
                await I.waitForFile(name);
            }
        }
        catch(error)
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");  

            throw new Error("waitForFile: Error while waiting for file to be present in current directory. "+name);
        }  
    }

    /**
     * Checks that file with a name including given text exists in the current directory.
     *
     *```js
    * z.handleDownloads();
    * z.click('Download as PDF');
    * z.amInPath('output/downloads');
    * z.seeFileNameMatching('.pdf');
    * ```
    */
    static async seeFileNameMatching(text: string)
    {
        try
        {
            await I.wait(2);
            await I.seeFileNameMatching(text);
        }
        catch(error)
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");  

            try
            {
                await I.wait(2);
                await I.seeFileNameMatching(text);
            }
            catch(error)
            {
                throw new Error("seeFileNameMatching: Error while checking that file with a name including given text exists in the current directory. "+text);
            }
        }  
    }

    /**
     * Checks that file found by `seeFile` doesn't include text.
     * @param {string} text
     * @param {string} [encoding='utf8']
     */
    static async dontSeeInThisFile(text: string, encoding?: string)
    {
        try
        {
            if(util.isNullOrUndefined(encoding))
            {
                await I.dontSeeInThisFile(text);
            }
            else
            {
                await I.dontSeeInThisFile(text, encoding);
            }
        }
        catch(error)
        {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");  

            try
            {
                if(util.isNullOrUndefined(encoding))
                {
                    await I.dontSeeInThisFile(text);
                }
                else
                {
                    await I.dontSeeInThisFile(text, encoding);
                }
            }
            catch(error)
            {
                throw new Error("dontSeeInThisFile: Error while checking that file found doesn't include text. "+text);
            }
        }    
    }

     /**
     * Checks that all elements with given locator have given attributes.
     *
     * ```js
     * z.seeAttributesOnElements('//form', { method: "post"});
     * ```
     *
     * @param {CodeceptJS.LocatorOrString} locator located by CSS|XPath|strict locator.
     * @param {object} attributes attributes and their values to check.
     *
     */
    static async seeAttributesOnElements(locator: string, attributes: any)
    {
        try
        {
            await I.wait(2);
            await I.seeAttributesOnElements(locator, attributes);
        }
        catch(error)
        {
            try 
            {
                let timestamp = new Date().getTime();
                await this.generateLog(timestamp, error);
                await I.saveScreenshot(timestamp.toString() + ".png"); 
    
                await I.wait(2);
                await I.seeAttributesOnElements(locator, attributes); 

            } catch (error)
            {
                throw new Error("seeAttributesOnElements: Error while checking that all elements with given locator have given attributes. "+locator);
            }
        }
    }

    /**
     * Checks that a given Element is present in the DOM
     * Element is located by CSS or XPath.
     *
     * ```js
     * z.seeElementInDOM('#modal');
     * ```
     * @param {string} locator element located by CSS|XPath|strict locator.
     *
     */
    static async seeElementInDOM(locator: string)
    {
        try
        {
            await I.seeElementInDOM(locator);
        } 
        catch (error) 
        {
            try 
            {
                let timestamp = new Date().getTime();
                await this.generateLog(timestamp, error);
                await I.saveScreenshot(timestamp.toString() + ".png"); 

                await I.seeElementInDOM(locator);
            } 
            catch (error) 
            {
                throw new Error("seeElementInDOM: Error while checking that a given Element is present in the DOM. "+locator);
            }
        }
    }

     /**
     * Appends text into the field.
     * A field can be located by text, accessibility id, id.
     *
     * ```js
     * z.appendField('name', 'davert');
     * ```
     *
     * @param {string} field
     * @param {string} value
     */
    static async appendField(locator: string, value: string)
    {
        try
        {
            await this.click(locator);

            await I.appendField(locator, value); 
        } 
        catch (error)
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
               
                this.click(locator);

                await I.appendField(locator, value);
                
                await this.generateLog(timestamp, error);
                
            } catch (error) 
            {
                throw new Error("appendField: Error while appending the text into the field.. "+locator);
            }
        }
    }

     /**
     * Waits for an element to become not attached to the DOM on a page (by default waits for 1sec).
     * Element can be located by CSS or XPath.
     *
     * ```js
     * z.waitForDetached('#popup');
     * ```
     *
     * @param {string} locator element located by CSS|XPath|strict locator.
     * @param {number} timeout (optional, `1` by default) time in seconds to wait
     */
    static async waitForDetached(locator: string, timeout?: number) {
        try {
            if (!util.isNullOrUndefined(timeout)) {
                await I.waitForDetached(locator, timeout);
            }
            else {
                await I.waitForDetached(locator);
            }
        }
        catch (error) {
            let timestamp = new Date().getTime();
            await this.generateLog(timestamp, error);
            await I.saveScreenshot(timestamp.toString() + ".png");            
            throw new Error("waitForDetached : Error while waiting for element to become not attached to the DOM located by : " + locator);

        }
    }

    /**
     * Sends POST request to API.
     *
     * ```js
     * I.sendPostRequest('/api/users.json', { "email": "user@user.com" });
     * ```
     *
     * @param {*} url
     * @param {*} payload
     * @param {object} headers
     */
    static async sendPostRequest(url: any, payload?: any, headers?: any)
    {
        let response: any;
        try
        {
           response = await I.sendPostRequest(url, payload, headers); 
        } 
        catch (error)
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
               
                response = await I.sendPostRequest(url, payload, headers);
                
                await this.generateLog(timestamp, error);
                
            } catch (error) 
            {
                throw new Error("sendPostRequest: Error while sending POST request to API.");
            }
        }
        return response;
    }

    /**
     * Send GET request to REST API
     *
     * ```js
     * I.sendGetRequest('/api/users.json');
     * ```
     *
     * @param {*} url
     * @param {object} headers
     */
    static async sendGetRequest(url: any, headers?: any)
    {
        let response: any;
        try
        {
           response = await I.sendGetRequest(url, headers); 
        } 
        catch (error)
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
               
                response = await I.sendGetRequest(url, headers);
                
                await this.generateLog(timestamp, error);
                
            } catch (error) 
            {
                throw new Error("sendGetRequest: Error while sending GET request to REST API");
            }
        }
        return response;
    }

     /**
     * Opposite to `seeElementInDOM`. Checks that element is not on page.
     *
     * ```js
     * I.dontSeeElementInDOM('.nav'); // checks that element is not on page visible or not
     * ```
     *
     * @param {string} locator located by CSS|XPath|Strict locator.
     */
    static async dontSeeElementInDOM(locator: string)
    {
        try
        {
           await I.dontSeeElementInDOM(locator); 
        } 
        catch (error)
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
               
                await I.dontSeeElementInDOM(locator);
                
                await this.generateLog(timestamp, error);
                
            } catch (error) 
            {
                throw new Error("dontSeeElementInDOM: Error while checking that element is not on page");
            }
        }
    }

    /**
     * Checks that text is equal to provided one.
     *
     * ```js
     * I.seeTextEquals('text', 'h1');
     * ```
     *
     * @param {string} text element value to check.
     * @param {string} optional [context]  element located by CSS|XPath|strict locator.
     */
    static async seeTextEquals(text: string, context?: string)
    {
        try
        {
          if(util.isNullOrUndefined(context))
          {
            await I.wait(2);  
            await I.seeTextEquals(text); 
          }
          else
          {
            await I.waitForVisible(context);
            await I.seeTextEquals(text, context); 
          }
        } 
        catch (error)
        {
            try 
            {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
               
                if(util.isNullOrUndefined(context))
                {
                  await I.wait(2);
                  await I.seeTextEquals(text); 
                }
                else
                {
                  await I.waitForVisible(context);
                  await I.seeTextEquals(text, context); 
                }

                await this.generateLog(timestamp, error);
                
            } catch (error) 
            {
                throw new Error("seeTextEquals: Error while checking that text is equal to provided one.");
            }
        }
    }

    /**
     * Checks that title is equal to provided one.
     *
     * ```js
     * z.seeTitleEquals('Test title.');
     * ```
     *
     * @param {string} text value to check.
     */
    static async  seeTitleEquals(text: string)
    {
        try
        {
            await I.seeTitleEquals(text);            
        } 
        catch (error) {
            let timestamp = new Date().getTime();
            await I.saveScreenshot(timestamp.toString() + ".png");
            await this.generateLog(timestamp, error);
            throw new Error("seeTitleEquals: Error while checking that title is equal to provided one.");
        }
    }

    /**
     * Clears a cookie by name,
     * if none provided clears all cookies.
     *
     * ```js
     * z.clearCookie();
     * z.clearCookie('test');
     * ```
     *
     * @param {string} (optional, `null` by default) cookie name
     * 
     */
       static async clearCookie(cookie?: string)
       {
            try {
                if(util.isNullOrUndefined(cookie))
                {
                    await I.clearCookie();
                }
                else
                {
                    await I.clearCookie(cookie);
                }
                
            } catch (error) {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
                await this.generateLog(timestamp, error);
                throw new Error("clearCookie: Error while clearing a cookie by name");
            }
       }

       /**
         * Retrieves the innerHTML from an element located by CSS or XPath and returns it to test.
         * Resumes test execution, so **should be used inside async function with `await`** operator.
         * If more than one element is found - an array of HTMLs returned.
         *
         * ```js
         * let postHTML = await z.grabHTMLFrom('#post');
         * ```
         *
         * @param {CodeceptJS.LocatorOrString} element located by CSS|XPath|strict locator.
         * @returns {Promise<string>} HTML code for an element
         */
        static async grabHTMLFrom(locator:string): Promise<string>
        {
            let postHTML:string;
            try {
                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    postHTML = await I.grabHTMLFrom(locator);
                }
                else {
                    postHTML = await I.grabHTMLFrom(locator);
                }               

            } catch (error) {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
                try 
                {
                    if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                    {
                        await I.waitForElement(locator);
                        await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                        await I.waitForVisible(locator);
                        postHTML = await I.grabHTMLFrom(locator);
                    }
                    else {
                        postHTML = await I.grabHTMLFrom(locator);
                    } 
                    await this.generateLog(timestamp, error);
                } 
                catch (error) {
                    throw new Error("grabHTMLFrom: Error while retrieving the innerHTML from an element");
                }               
               
            }
            return postHTML;
        }
        
        /**
         * Retrieves all the innerHTML from elements located by CSS or XPath and returns it to test. 
         * Resumes test execution, so should be used inside async function with await operator.
         *
         * ```js
         * let postHTML = await z.grabHTMLFromAll('#post');
         * ```
         *
         * @param {CodeceptJS.LocatorOrString} element located by CSS|XPath|strict locator.
         * @returns {Promise<string>} HTML code for an element
         */
        static async grabHTMLFromAll(locator:string): Promise<any>
        {
            let postHTMLs:any[]=[];
            try 
            {              
                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    postHTMLs  = await I.grabHTMLFromAll(locator);
                }
                else {
                    postHTMLs  = await I.grabHTMLFromAll(locator);
                }             
                                
            } 
            catch (error) {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
                try 
                {
                    if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                    {
                        await I.waitForElement(locator);
                        await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                        await I.waitForVisible(locator);
                        postHTMLs  = await I.grabHTMLFromAll(locator);
                    }
                    else {
                        postHTMLs  = await I.grabHTMLFromAll(locator);
                    } 
                    await this.generateLog(timestamp, error); 
                } 
                catch (error) {
                    throw new Error("grabHTMLFromAll: Error while retrieving all the innerHTML from elements");
                }
               
            }
            return postHTMLs;
        }

         /**
         * Retrieves the cssProperty from an element located by CSS or XPath and returns it to test.
         * Resumes test execution, so **should be used inside async function with `await`** operator.
         * If more than one element is found - an array of HTMLs returned.
         *
         * ```js
         * const value = await I.grabCssPropertyFrom('h3', 'font-weight');
         * ```
         *
         * @param {CodeceptJS.LocatorOrString} element located by CSS|XPath|strict locator.
         *  @param {CodeceptJS.String} cssProperty CSS property name.
         * @returns {Promise<string>} CSS value
         */
        static async grabCssPropertyFrom(locator:string,cssProperty:string): Promise<string>
        {
            let postHTML:string='';
            try {
                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    postHTML = await I.grabCssPropertyFrom(locator,cssProperty);
                }
                else {
                    postHTML = await I.grabCssPropertyFrom(locator,cssProperty);
                }               

            } catch (error) {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
                try 
                {
                    if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                    {
                        await I.waitForElement(locator);
                        await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                        await I.waitForVisible(locator);
                        postHTML = await I.grabCssPropertyFrom(locator,cssProperty);
                    }
                    else {
                        postHTML = await I.grabCssPropertyFrom(locator,cssProperty);
                    } 
                    await this.generateLog(timestamp, error);
                } 
                catch (error) {
                    throw new Error("grabCssPropertyFrom: Error while retrieving the css property from an element");
                }               
               
            }
            return postHTML;
        }

        static async grabCssPropertyFromAll(locator:string,cssProperty:string): Promise<any>
        {
            let postHTML:any[]=[];
            try {
                if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                {
                    await I.waitForElement(locator);
                    await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                    await I.waitForVisible(locator);
                    postHTML = await I.grabCssPropertyFromAll(locator,cssProperty);
                }
                else {
                    postHTML = await I.grabCssPropertyFromAll(locator,cssProperty);
                }               

            } catch (error) {
                let timestamp = new Date().getTime();
                await I.saveScreenshot(timestamp.toString() + ".png");
                try 
                {
                    if (locator.startsWith("/") || locator.startsWith(".") || locator.startsWith("#") || locator.startsWith("{") || locator.startsWith("(")) 
                    {
                        await I.waitForElement(locator);
                        await I.scrollIntoView(locator,{behavior: "auto", block: "center", inline: "center"});
                        await I.waitForVisible(locator);
                        postHTML = await I.grabCssPropertyFromAll(locator,cssProperty);
                    }
                    else {
                        postHTML = await I.grabCssPropertyFromAll(locator,cssProperty);
                    } 
                    await this.generateLog(timestamp, error);
                } 
                catch (error) {
                    throw new Error("grabCssPropertyFrom: Error while retrieving the css property from an element");
                }               
               
            }
            return postHTML;
        }

    static async generateLog(timestamp: number, error: any) {
        console.log("failed for " + error.stack);
        // await I.saveScreenshot(timestamp.toString() + ".png");
    }

    /**
     * Checks if element is visible and returns true of false accordingly.
     *
     * ```js
     * const value = await z.checkIfVisible("//input[@id='123']");
     * ```
     *
     * @param {string|object} locator element located by CSS|XPath|strict locator.
     * @returns {boolean} boolean
     */
    static async checkIfVisible(selector :any, ...options : any): Promise<any> {
        let flag = false;
        try {
            if(await I.grabNumberOfVisibleElements(selector)) {
                flag = true;
            }
            else {
                flag = false;
            }
        }
        catch (err) {
            console.log(err);
        }
        return flag;
    }

    /**
     * Downloads the file from docker agent to code base agent.
     * We can also pass an expected file name to be set for the downloaded file.
     *
     * ```js
     * const fileName = await z.dockerFileDownload();
     * const fileName = await z.dockerFileDownload("test.pdf");
     * ```
     *
     * @returns {string} file path in the repo (./tempDownload/test.pdf)
     */
    static async dockerFileDownload(expectedFileName?: string){
        const config = require('codeceptjs').config.get();
        const ggrPort=config.helpers.WebDriver.port;
        const ggrHost=config.helpers.WebDriver.host;
        
        var sessions = this.prototype.helpers['WebDriver'].browser; 
        console.log("Browser session id ",sessions.sessionId)
    
        const helper = this.prototype.helpers['REST'];
    
        //SELENOID_HOST - ggr machien ip and port
        let res=await helper.sendGetRequest(`http://${ggrHost}:${ggrPort}/host/${sessions.sessionId}`);
        console.log(`Response from ggr machine api\n${JSON.stringify(res.data)}`);
    
        let ipAdress= res.data["Name"]
        let port= res.data["Port"]
        console.log("Selenoid IP address "+ipAdress)

        //ipAdress - selenoid machien ip and port
        let response = await helper.sendGetRequest(`http://${ipAdress}:${port}/status`);
        console.log(`Response from status api\n${JSON.stringify(response.data)}`);
        
        
        let obj1 = response.data["browsers"]["chrome"];
    
        let sessionId = "";
        for (const [key, value] of Object.entries(obj1)) {
            if(Object.keys(value as any).length) {
                console.log(`key found --> ${key}`);
                let sessionArray = (value as any)["test"]["sessions"];
                for (const element of sessionArray) {
                sessionId = element["id"];
                    if(sessions.sessionId.includes(sessionId)) {
                        console.log(`id --> ${sessionId}`);
                        break;
                    }
                }
                break;
            }
        } 
        console.log(`Session id retrieved\n${sessionId}`);
    
        let  secresponse = await helper.sendGetRequest(`http://${ipAdress}:${port}/download/${sessionId}`);
        console.log(`Response from session api\n${JSON.stringify( secresponse.data)}`);
    
        let defaultCxmlDocument = new DOMParser().parseFromString(secresponse.data, "application/xml");

        ///////////////To get latest fiel downloaded///////////
        let fileIndex = (defaultCxmlDocument.documentElement.getElementsByTagName("a").length-1);
        let fileName = defaultCxmlDocument.documentElement.getElementsByTagName("a")[fileIndex].getAttribute("href");
        
        /////////// To get first file downloaded(we can pass index instead of zero)
        // let fileName = defaultCxmlDocument.documentElement.getElementsByTagName("a")[0].getAttribute("href");
        console.log(`File name\n${fileName}`);
    
        let newFileName = "./tempDownload/"+fileName; 
        if(expectedFileName) {
            newFileName = "./tempDownload/" + expectedFileName;
        }
        let file = fs.createWriteStream(newFileName); 
        // const file = fs.createWriteStream("./tempDownload/"+fileName); 
        await new Promise(function(resolve, reject){
            http.get(`http://${ipAdress}:${port}/download/${sessionId}/${fileName}`, (response) => {
                response.pipe(file);
                setTimeout(() => {
                    resolve(true);
                }, 10000);
            });
        });
        return "./tempDownload/"+fileName;
    }

    /**
     * Starts the browser explicitly.
     *
     * ```js
     * await z.startBrowser();
     * ```
     */
    static async startBrowser() {
        try {
            console.log("========================>Start the Browser");
            await this.prototype.helpers["WebDriver"]._startBrowser();
        }
        catch (err) {
            console.log("========================>ERROR"+err);
            console.log("========================>Starting the Browser Failed...");
        }
    }

    /**
     * Stops the browser explicitly.
     *
     * ```js
     * await z.stopBrowser();
     * ```
     */
    static async stopBrowser() {
        try {
            console.log("========================>Stop the Browser");
            await this.prototype.helpers["WebDriver"]._stopBrowser();
        }
        catch (err) {
            console.log("========================>ERROR"+err);
            console.log("========================>Stopping the Browser Failed...");
        }    
    }

    static async grabPropertyOrAttributeValue(locator: string, attribute: string, elementLocatorStrategy: string = 'xpath') {
        let value = '';
        return new Promise(async (resolve, reject) => {
            try {
                const browser = await this.prototype.helpers['WebDriver'].browser;
                const elementRef = await browser.findElement(elementLocatorStrategy, locator);
                Object.keys(elementRef)
                    .forEach(async function eachKey(key) {
                        value = await browser.getElementAttribute(elementRef[key], attribute);
                        if (value == null) {
                            value = await browser.getElementProperty(elementRef[key], attribute);
                        }
                        resolve(value);
                    });
            }
            catch (error) {
                throw new Error('Error: Grab Property and attribute is throwing error for locator : '+locator + 'while fetching attribute or property : '+attribute); 
                }
        })
    }

    static async grabPropertyOrAttributeValueFromAll(locator: string, attribute: string, elementLocatorStrategy: string = 'xpath') {
        let value:any;
        let valueArray:any=[];
        let count=0;
        return new Promise(async (resolve, reject) => {
            try {
                const browser = await this.prototype.helpers['WebDriver'].browser;
                const elementRef = await browser.findElements(elementLocatorStrategy, locator);
                for(const element in elementRef){
                Object.keys(elementRef[element])
                    .forEach(async function eachKey(key) {
                        value = await browser.getElementAttribute(elementRef[element][key], attribute);
                        if (value == null) {
                            value = await browser.getElementProperty(elementRef[element][key], attribute);
                        }
                        count=count+1;
                        valueArray.push(value);
                    if (count == elementRef.length) {
                        resolve(valueArray);
                    }
                    })
                    
                }
            }
            catch (error) {
                reject(error);
                throw new Error('Error: Grab attribute from all is throwing error for locator : '+locator + 'while fetching attribute or property : '+attribute); 
                }
        })
    }
}
