import { autoHeal } from "./autoHeal";
import { nonHeal } from "./nonHeal";
const util = require('util');
const { I } = inject();

export class z extends Helper {

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then performs click action.
     * Throws error if any of the action fails in two attempts.
     * The second parameter is a context (CSS or XPath locator) to narrow the search.
     * ```js
     * // simple link
     * z.click('Logout');
     * // button of form
     * z.click('Submit');
     * // CSS button
     * z.click('#form input[type=submit]');
     * // XPath
     * z.click('//form/*[@type=submit]');
     * // link in context
     * z.click('Logout', '#nav');
     * // using strict locator
     * z.click({css: 'nav a.login'});
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string} context (optional, null by default) element to search in CSS|XPath|Strict locator.
     */
    static async click(locator: any, ElementContext?: any) {
        let context;
        if (Array.isArray(ElementContext))
            context = ElementContext[0]
        else
            context = ElementContext
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("("))) {
                await autoHeal.click(locator[0], locator[1], 120, 0, context);
            }
            else
                await nonHeal.click(locator[0], context)
        }
        else
            await nonHeal.click(locator, context)
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & enabled, then clears field value.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.clearField('Email');
     * z.clearField('user[email]');
     * z.clearField('#email');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     */
    static async clearField(locator: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("("))) {
                await autoHeal.clearField(locator[0], locator[1]);
            }
            else
                await nonHeal.clearField(locator[0]);
        }
        else
            await nonHeal.clearField(locator);
    }


    /**
     * Waits for element to be visible, Clears element value & fills given value.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * // by label
     * z.fillField('Email', 'hello@world.com');
     * // by name
     * z.fillField('password', secret('123456'));
     * // by CSS
     * z.fillField('form#login input[name=username]', 'John');
     * // or by strict locator
     * z.fillField({css: 'form#login input[name=username]'} 'John');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string} value
     */
    static async fillField(locator: any, value: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("("))) {
                await autoHeal.fillField(locator[0], value, locator[1]);
            }
            else
                await nonHeal.fillField(locator[0], value)
        }
        else
            await nonHeal.fillField(locator, value)
    }


    /**
     * Checks for the visibility of text, if optional context parameter given then searches for the visibility of text on the locator element.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.see('Welcome'); // text welcome on a page
     * z.see('Welcome', '.content'); // text inside .content div
     * z.see('Register', {css: 'form.register'}); // use strict locator
     * ```
     * @param {string} text
     * @param {string} context optional, element located by CSS|Xpath|strict locator in which to search for text.
     */
    static async see(text: any, context?: any) {
            if (Array.isArray(context)) 
            await nonHeal.see(text, context[0]);
            else
            await nonHeal.see(text, context);
    }

    /**
     * Checks the visibilty of an element.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.seeElement('#modal');
     * ```
     * @param {string} locator CSS|XPath|strict locator. 
     */
    static async seeElement(locator: any) {
        if (Array.isArray(locator))
            await autoHeal.seeElement(locator[0],locator[1]);
        else
            await nonHeal.seeElement(locator);
    }

    /**
     * Opens a webpage in browser with given url.
     * ```js
     * z.amOnPage('/'); // opens main page of website
     * z.amOnPage('https://github.com'); // opens github
     * z.amOnPage('/login'); // opens a login page
     * ```
     * @param {string} url 
     */
    static async amOnPage(url: string) {
        await nonHeal.amOnPage(url);
    }

    /**
     * Asserts that two values are equal. If they are not,an assertion error is thrown. 
     * ```js
     * z.assertEqual(expected,actual)
     * ```
     * @param {any} expected 
     * @param {any} actual 
     */
    static async assertEqual(expected: any, actual: any, msg?: string) {
        await nonHeal.assertEqual(expected, actual, msg)
    }

    /**
     * Returns all window handles. Useful for referencing a specific handle when calling I.switchToWindow(handle)
     * ```js
     * const windows = await I.grabAllWindowHandles();
     * ```
     * @returns {Array} windows
     */
    static async grabAllWindowHandles(): Promise<any> {
        return await nonHeal.grabAllWindowHandles();
    }

    /**
     * Returns current window handle
     * ```js
     * const window = await z.grabCurrentWindowHandle();
     * ```
     * @returns {any} window
     */
    static async grabCurrentWindowHandle(): Promise<any> {
        return await nonHeal.grabCurrentWindowHandle();
    }

    /**
     * Grabs the number of visible elements
     * ```js
     * let numOfElements = await z.grabNumberOfVisibleElements('p');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @returns {any} totalElements
     */
    static async grabNumberOfVisibleElements(locator: any): Promise<any> {
        if (Array.isArray(locator))
            return await nonHeal.grabNumberOfVisibleElements(locator[0]);
        else
            return await nonHeal.grabNumberOfVisibleElements(locator);
    }

    /**
     * Open new tab and switch to it.
     * ```js
     * z.openNewTab();
     * ```
     */
    static async openNewTab() {
        await nonHeal.openNewTab();
    }

    /**
     * Presses a key or a key in combination with modifier keys in the browser (on a focused element).
     * ```js
    * z.pressKey('Backspace');
    * ```
    *
    * To press a key in combination with modifier keys, pass the sequence as an array. All modifier keys (`'Alt'`, `'Control'`, `'Meta'`, `'Shift'`) will be released afterwards.
    *
    * ```js
    * z.pressKey(['Control', 'Z']);
    * ```
    *
    * For specifying operation modifier key based on operating system it is suggested to use `'CommandOrControl'`.
    * This will press `'Command'` (also known as `'Meta'`) on macOS machines and `'Control'` on non-macOS machines.
    *
    * ```js
    * z.pressKey(['CommandOrControl', 'Z']);
    * ```
    *
    * Some of the supported key names are:
    * - `'AltLeft'` or `'Alt'`
    * - `'AltRight'`
    * - `'ArrowDown'`
    * - `'ArrowLeft'`
    * - `'ArrowRight'`
    * - `'ArrowUp'`
    * - `'Backspace'`
    * - `'Clear'`
    * - `'ControlLeft'` or `'Control'`
    * - `'ControlRight'`
    * - `'Command'`
    * - `'CommandOrControl'`
    * - `'Delete'`
    * - `'End'`
    * - `'Enter'`
    * - `'Escape'`
    * - `'F1'` to `'F12'`
    * - `'Home'`
    * - `'Insert'`
    * - `'MetaLeft'` or `'Meta'`
    * - `'MetaRight'`
    * - `'Numpad0'` to `'Numpad9'`
    * - `'NumpadAdd'`
    * - `'NumpadDecimal'`
    * - `'NumpadDivide'`
    * - `'NumpadMultiply'`
    * - `'NumpadSubtract'`
    * - `'PageDown'`
    * - `'PageUp'`
    * - `'Pause'`
    * - `'Return'`
    * - `'ShiftLeft'` or `'Shift'`
    * - `'ShiftRight'`
    * - `'Space'`
    * - `'Tab'`
     * @param {string|Array} key
     */
    static async pressKey(key: string | string[]) {
        await nonHeal.pressKey(key);
    }

    /**
     * Reload the current page.
     * ```js
     * z.refreshPage();
     * ```
     */
    static async refreshPage() {
        await nonHeal.refreshPage();
    }

    /**
     * Saves a screenshot to ouput folder (set in codecept.json or codecept.conf.js). 
     * Filename is relative to output folder. Optionally resize the window to the full available page scrollHeight and scrollWidth
     * to capture the entire page by passing true in as the second argument.
     * ```js
     * z.saveScreenshot('debug.png');
     * z.saveScreenshot('debug.png', true) //resizes to available scrollHeight and scrollWidth before taking screenshot
     * ``` 
     * @param {string} fileName
     * @param {boolean} fullPage optional
     */
    static async saveScreenshot(fileName: string, fullPage?: boolean) {
        await nonHeal.saveScreenshot(fileName, fullPage);
    }

    /**
     * Switches frame or in case of null locator reverts to parent.
     * ```js
     * z.switchTo('iframe'); // switch to first iframe
     * z.switchTo(); // switch back to main page
     * ```
     * @param {string} locator optional, CSS|XPath|strict locator.
     */
    static async switchTo(locator?: any) {
        if (Array.isArray(locator))
            await nonHeal.switchTo(locator[0]);
        else
            await nonHeal.switchTo(locator);
    }

    /**
     * Switch focus to a particular tab by its number. It waits tabs loading and then switch tab.
     * ```js
     * z.switchToNextTab();
     * z.switchToNextTab(2);
     * ```
     * @param {number} tabNum optional, number of tabs to switch forward, default: 1.
     * @param {number} timeout optional, time in seconds to wait.
     */
    static async switchToNextTab(tabNum?: number, timeout?: number) {
        await nonHeal.switchToNextTab(tabNum, timeout);
    }

    /**
     * Switch to the window with a specified handle.
     * ```js
     * const windows = await z.grabAllWindowHandles();
     * //  do something
     * await z.switchToWindow( windows[0] );
     *
     * const window = await z.grabCurrentWindowHandle();
     * //  do something
     * await z.switchToWindow( window );
     * ``` 
     * @param {string} window
     */
    static async switchToWindow(window: any) {
        await nonHeal.switchToWindow(window);
    }

    /**
     * Pauses execution for a number of seconds.
     * ```js
     * z.wait(2); // wait 2 secs
     * ```
     * @param {number} timeout
     */
    static async wait(timeout: any) {
        await nonHeal.wait(timeout);
    }

    /**
     * Waits for the visibility of element & for the specified value to be in value attribute.
     * ```js
     * z.waitForValue('//input', "GoodValue");
     * ```
     * @param {string} locator input field locator.
     * @param {string} value
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     */
    static async waitForValue(locator: any, value: string, timeout: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.waitForValue(locator[0], value, locator[1], timeout);
            else
                await nonHeal.waitForValue(locator[0], value, timeout);
        }
        else
            await nonHeal.waitForValue(locator, value, timeout);
    }

    /**
     * Waits for the visibility of element.
     * ```js
     * z.waitForVisible('#popup');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"  
     */
    static async waitForVisible(locator: string, timeout: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.waitForVisible(locator[0], locator[1], timeout);
            else
                await nonHeal.waitForVisible(locator[0], timeout);
        }
        else
            await nonHeal.waitForVisible(locator, timeout);
    }

    /**
     * Waits for a text to appear. 
     * Element can be located by CSS or XPath. Narrow down search results by providing context.
     * ```js
     * z.waitForText('Thank you, form has been submitted');
     * z.waitForText('Thank you, form has been submitted', 5, '#modal');
     * ```
     * @param {string} text
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     * @param {string} context optional, element located by CSS|XPath|strict locator.
     
     */
    static async waitForText(text: string, timeout?: number, context?: any) {
        if (Array.isArray(context)) 
        await nonHeal.waitForText(text, timeout, context[0]);
        else
        await nonHeal.waitForText(text, timeout, context);
    }


    /**
     * Waits for an element to be removed or become invisible on a page. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForInvisible('#popup');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     */
    static async waitForInvisible(locator: any, timeout: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.waitForInvisible(locator[0], locator[1], timeout);
            else
                await nonHeal.waitForInvisible(locator[0], timeout);
        }
        else
            await nonHeal.waitForInvisible(locator, timeout);
    }

    /**
     * Waits for an element to become enabled. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForEnabled('#popup');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"
     */
    static async waitForEnabled(locator: any, timeout: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.waitForEnabled(locator[0], locator[1], timeout);
            else
                await nonHeal.waitForEnabled(locator[0], timeout);
        }
        else
            await nonHeal.waitForEnabled(locator, timeout);
    }

    /**
     * Waits for an element to be present on page. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForElement('.btn.continue');
     * z.waitForElement('.btn.continue', 5); // wait for 5 secs
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout" 
     */
    static async waitForElement(locator: any, timeout: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.waitForElement(locator[0], locator[1], timeout);
            else
                await nonHeal.waitForElement(locator[0], timeout);
        }
        else
            await nonHeal.waitForElement(locator, timeout);
    }

    /**
     * Waits for an element to be clickable. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.waitForClickable('.btn.continue');
     * z.waitForClickable('.btn.continue', 5); // wait for 5 secs
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {number} timeout optional, default timeout is value mentioned against key "waitForTimeout"    
     */
    static async waitForClickable(locator: any, timeout: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.waitForClickable(locator[0], locator[1], timeout);
            else
                await nonHeal.waitForClickable(locator[0], timeout);
        }
        else
            await nonHeal.waitForClickable(locator, timeout);
    }

    /**
     * Waits for an element to be visible. 
     * Element can be located by CSS or XPath.
     * ```js
     * z.attachFile('Avatar', 'data/avatar.jpg');
     * z.attachFile('form input[name=avatar]', 'data/avatar.jpg');
     * ```
     * @param {string} locator label|name|CSS|XPath|strict locator.
     * @param {string} filepath local file path relative to codecept.json config file
     */
    static async attachFile(locator: any, filepath: string) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.attachFile(locator[0], filepath, locator[1]);
            else
                await nonHeal.attachFile(locator[0], filepath);
        }
        else
            await nonHeal.attachFile(locator, filepath);
    }

    /**
     * Switch focus to a particular tab by its number. It waits tabs loading and then switch tab.
     * ```js
     * z.switchToPreviousTab();
     * z.switchToPreviousTab(2);
     * ```
     * @param {number} tabNum optional, number of tabs to switch backward, default: 1.
     * @param {number} timeout optional, time in seconds to wait.
     */
    static async switchToPreviousTab(tabNum?: number, timeout?: number) {
        await nonHeal.switchToPreviousTab(tabNum, timeout);
    }

    /**
     * Close current tab.
     * ```js
     * z.closeCurrentTab();
     * ```
     */
    static async closeCurrentTab() {
        await nonHeal.closeCurrentTab();
    }

    /**
     * Close all tabs except for the current one.
     * ```js
     * z.closeOtherTabs();
     * ```
     */
    static async closeOtherTabs() {
        await nonHeal.closeOtherTabs();
    }

    /**
     * Scroll element into viewport.
     * ```js
     * z.scrollIntoView('#submit');
     * z.scrollIntoView('#submit', { behavior: "smooth", block: "center", inline: "center" });
     * z.scrollIntoView('#submit', true);
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {any} alignTop optional, or scrollIntoViewOptions (optional), 
     * see https://developer.mozilla.org/en-US/docs/Web/API/Element/scrollIntoView.
     
     */
    static async scrollIntoView(locator: any, alignTop?: any) {
        if (Array.isArray(locator))
            await nonHeal.scrollIntoView(locator[0], alignTop);
        else
            await nonHeal.scrollIntoView(locator, alignTop);
    }

    /**
     * Retrieves a text from an element located by CSS or XPath and returns it to test. 
     * Resumes test execution, so should be used inside static async with await operator.
     * If multiple elements found returns an array of texts.
     * ```js
     * let pin = await z.grabTextFrom('#pin');
     * ```
     * @param {string} locator element located by CSS|XPath|strict locator.
     * @returns {string|string[]} text 
     */
    static async grabTextFrom(locator: any): Promise<any> {
        let text: string = '';
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                text = await autoHeal.grabTextFrom(locator[0], locator[1]);
            else
                text = await nonHeal.grabTextFrom(locator[0]);
        }
        else {
            text = await nonHeal.grabTextFrom(locator);
        }
        return text;
    }

    static async grabTextFromAll(locator: any): Promise<any> {
        let text: any[] = [];
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                text = await autoHeal.grabTextFromAll(locator[0], locator[1]);
            else
                text = await nonHeal.grabTextFromAll(locator[0]);
        }
        else
            text = await nonHeal.grabTextFromAll(locator);
        return text;
    }

    /**
     * Retrieves an attribute from an element located by CSS or XPath and returns it to test. An array as a result will be returned if there are more than one matched element. 
     * Resumes test execution, so should be used inside static async with await operator.
     * ```js
     * let hint = await z.grabAttributeFrom('#tooltip', 'title');
     * ```
     * @param {string} locator element located by CSS|XPath|strict locator.
     * @param {string} attribute attribute name.
     * @returns {string} text 
     */
    static async grabAttributeFrom(locator: any, attribute: string): Promise<any> {
        let value: any = "";
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                value = await autoHeal.grabAttributeFrom(locator[0], attribute, locator[1]);
            else
                value = await nonHeal.grabAttributeFrom(locator[0], attribute);
        }
        else
            value = await nonHeal.grabAttributeFrom(locator, attribute);
        return value;
    }

    static async grabAttributeFromAll(locator: any, attribute: string): Promise<any> {
        let value: any = [];
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                value = await autoHeal.grabAttributeFromAll(locator[0], attribute, locator[1]);
            else
                value = await nonHeal.grabAttributeFromAll(locator[0], attribute);
        }
        else
            value = await nonHeal.grabAttributeFromAll(locator, attribute);
        return value;
    }


    /**
         * Scrolls element into viewport, waits for that element to become visible & clickable. Also checks if element is already checked.
         * Then selects a checkbox or radio button.
         * 
         * The second parameter is a context (CSS or XPath locator) to narrow the search.
         * ```js
         * z.checkOption('#agree');
         * z.checkOption('I Agree to Terms and Conditions');
         * z.checkOption('agree', '//form');
         * ```
         * @param {string} locator checkbox located by label | name | CSS | XPath | strict locator.
         * @param {string} context optional element located by CSS | XPath | strict locator.
         */
    static async checkOption(locator: any, context?: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.checkOption(locator[0], locator[1], 120, 0, context);
            else
            {
                if(Array.isArray(context))
                await nonHeal.checkOption(locator[0], context[0]);
                else
                await nonHeal.checkOption(locator[0], context);
            }
        }
        else
            {
                if(Array.isArray(context))
                await nonHeal.checkOption(locator, context[0]);
                else
                await nonHeal.checkOption(locator, context);
            }
    }

    /**
     * Checks that a text is not present on a page, if optional context parameter given then searches for the invisibility of text on the locator element.
     * Use context parameter to narrow down the search.
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.dontSee('Login'); // assume we are already logged in.
     * z.dontSee('Login', '.nav'); // no login inside .nav element
     * ```
     * @param {string} text which is not present.
     * @param {string} context optional, element located by CSS|Xpath|strict locator in which to search for text.
     */
    static async dontSee(text: string, context?: any) {
            if (Array.isArray(context)) 
            await nonHeal.dontSee(text, context[0]);
            else
            await nonHeal.dontSee(text, context);
    }

    /**
     * Checks that element is not visible (or in DOM)
     * Throws error if any of the action fails in two attempts.
     * ```js
     * z.seeElement('#modal');
     * ```
     * @param {string} locator CSS|XPath|strict locator. 
     */
    static async dontSeeElement(locator: any) {
        if (Array.isArray(locator))
            await nonHeal.dontSeeElement(locator[0]);
        else
            await nonHeal.dontSeeElement(locator);
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then performs double click action.
     * Throws error if any of the action fails in two attempts.
     * The second parameter is a context (CSS or XPath locator) to narrow the search.
     * ```js
     * z.doubleClick('Edit');
     * z.doubleClick('Edit', '.actions');
     * z.doubleClick({css: 'button.accept'});
     * z.doubleClick('.btn.edit');
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string} context (optional, null by default) element to search in CSS|XPath|Strict locator.
     */
    static async doubleClick(locator: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.doubleClick(locator[0], locator[1]);
            else
                await nonHeal.doubleClick(locator[0]);
        }
        else
            await nonHeal.doubleClick(locator);
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then Drag an item to a destination element.
     * Throws error if any of the action fails in two attempts.
     * 
     * ```js
     * z.dragAndDrop('#dragHandle', '#container');
     * ```
     * @param {string} srcLocator CSS|XPath|strict locator.
     * @param {string} destLocator CSS|XPath|Strict locator.
     */
    static async dragAndDrop(srcLocator: any, destLocator: any) {
        if (Array.isArray(srcLocator) && Array.isArray(destLocator))
            await nonHeal.dragAndDrop(srcLocator[0], destLocator[0]);
        else if (Array.isArray(srcLocator) && !Array.isArray(destLocator))
            await nonHeal.dragAndDrop(srcLocator[0], destLocator);
        else if (!Array.isArray(srcLocator) && Array.isArray(destLocator))
            await nonHeal.dragAndDrop(srcLocator, destLocator[0]);
        else
            await nonHeal.dragAndDrop(srcLocator, destLocator);
    }

    /**
     * Executes sync script on a page.
     * Pass arguments to function as additional parameters.
     * Will return execution result to a test.
     * In this case you should use static async function and await to receive results.
     *
     * Example with jQuery DatePicker:
     *
     * ```js
     * // change date of jQuery DatePicker
     * z.executeScript(function() {
     *   // now we are inside browser context
     *   $('date').datetimepicker('setDate', new Date());
     * });
     * ```
     * Can return values. Don't forget to use `await` to get them.
     *
     * ```js
     * let date = await z.executeScript(function(el) 
     *   // only basic types can be returned
     *   return $(el).datetimepicker('getDate').tostring();
     * } '#date'); // passing jquery selector
     * ```
     * @param {string|function} fn function to be executed in browser context.
     * @param {...any} args to be passed to function.     
     * */
    static async executeScript(fn: any | ((...params: any[]) => any), ...args: any[]): Promise<any> {
        let retunrtype = await I.executeScript(fn, args);
        return retunrtype;
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & clickable, then selects an option in a drop-down select.
     * Throws error if any of the action fails in two attempts.
     * 
     * ```js
     * z.selectOption('Choose Plan', 'Monthly'); // select by label
     * z.selectOption('subscription', 'Monthly'); // match option by text
     * z.selectOption('subscription', '0'); // or by value
     * z.selectOption('//form/select[@name=account]','Premium');
     * z.selectOption('form select[name=account]', 'Premium');
     * z.selectOption({css: 'form select[name=account]'} 'Premium');
     * ```
     * Provide an array for the second argument to select multiple options.
     * ```js
     * z.selectOption('Which OS do you use?', ['Android', 'iOS']);
     * ```
     * @param {string} locator CSS|XPath|strict locator.
     * @param {string|string[]} option CSS|XPath|Strict locator.
     */
    static async selectOption(locator: any, option: string) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.selectOption(locator[0], option, locator[1]);
            else
                await nonHeal.selectOption(locator[0], option);
        }
        else
            await nonHeal.selectOption(locator, option);
    }


    /**
     * Checks for element to be present, scrolls into view & Moves cursor to element matched by locator.
     * Extra shift can be set with offsetX and offsetY options.
     *
     * ```js
     * z.moveCursorTo('.tooltip');
     * z.moveCursorTo('#submit', 5,5);
     * ```
     *
     * @param {CodeceptJS.LocatorOrString} locator located by CSS|XPath|strict locator.
     * @param {number} [offsetX=0] (optional, `0` by default) X-axis offset.
     * @param {number} [offsetY=0] (optional, `0` by default) Y-axis offset.
     *
     *
     */
    static async moveCursorTo(locator: any, offsetX?: number, offsetY?: number) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.moveCursorTo(locator[0], locator[1], offsetX, offsetY);
            else
                await nonHeal.moveCursorTo(locator[0], offsetX, offsetY);
        }
        else
            await nonHeal.moveCursorTo(locator, offsetX, offsetY);
    }

    /**
     * Checks for element to be present, scrolls into view, waits for element to be visible and retrieves a value from a form element located by CSS or XPath, returns it to test.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let email = await z.grabValueFrom('input[name=email]');
     * ```
     * @param {CodeceptJS.LocatorOrString} locator field located by label|name|CSS|XPath|strict locator.
     * @returns {Promise<any>} attribute value
     *
     */
    static async grabValueFrom(locator: any): Promise<any> {
        let text: any = "";
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                text = await autoHeal.grabValueFrom(locator[0], locator[1]);
            else
                text = await nonHeal.grabValueFrom(locator[0]);
        }
        else
            text = await nonHeal.grabValueFrom(locator);
        return text;
    }

    static async grabValueFromAll(locator: any): Promise<any> {
        let text: any;
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                text = await autoHeal.grabValueFromAll(locator[0], locator[1]);
            else
                text = await nonHeal.grabValueFromAll(locator[0]);
        }
        else
            text = await nonHeal.grabValueFromAll(locator);
        return text;
    }

    /**
     * Dismisses the active JavaScript popup, as created by window.alert|window.confirm|window.prompt.
     * ```js
     * await z.cancelPopup();
     * ```
     */
    static async cancelPopup() {
        await nonHeal.cancelPopup();
    }

    /**
     * Accepts the active JavaScript native popup window, as created by window.alert|window.confirm|window.prompt.
     * Don't confuse popups with modal windows, as created by [various
     * libraries](http://jster.net/category/windows-modals-popups).
     * ```js
     * await z.acceptPopup();
     * ```
     */
    static async acceptPopup() {
        await nonHeal.acceptPopup();
    }

    /**
     * Checks that the active JavaScript popup, as created by `window.alert|window.confirm|window.prompt`, contains the
     * given string.
     *
     * @param {string} text value to check.
     * 
     * ```js
     * await z.seeInPopup("value");
     * ```
     * 
     */
    static async seeInPopup(text: string) {
        await nonHeal.seeInPopup(text);
    }

    /**
     * Gets a cookie object by name.
     * If none provided gets all cookies.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let cookie = await z.grabCookie('auth');
     * assert(cookie.value, '123456');
     * ```
     *
     * @param {?string} [name=null] cookie name.
     * @returns {Promise<string>} attribute value
     *
     */
    static async grabCookie(name: string): Promise<any> {
        return await nonHeal.grabCookie(name);
    }

    /**
     * Get current URL from browser
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let url = await z.grabCurrentUrl();
     * ```
     *
     */

    static async grabCurrentUrl(): Promise<any> {
        return await nonHeal.grabCurrentUrl();
    }

    /**
    * Grab number of open tabs.
    * Resumes test execution, so **should be used inside async function with `await`** operator.
    *
    * ```js
    * let tabs = await z.grabNumberOfOpenTabs();
    * ```
    *
    */

    static async grabNumberOfOpenTabs(): Promise<any> {
        return await nonHeal.grabNumberOfOpenTabs();
    }

    /**
     * Checks that current url contains a provided fragment.
     *
     * @param {string} url value to check.
     * 
     * ```js
     * await z.seeInCurrentUrl("url");
     * ```
     * 
     */
    static async seeInCurrentUrl(url: string) {
        await nonHeal.seeInCurrentUrl(url);
    }

    /**
     * Retrieves a page title and returns it to test.
     * Resumes test execution, so **should be used inside async function with `await`** operator.
     *
     * ```js
     * let tabs = await z.grabTitle();
     * ```
     *
     */

    static async grabTitle(): Promise<any> {
        return await nonHeal.grabTitle();
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible.
     * Then Verifies that the specified checkbox is not checked.
     * 
     * ```js
     * z.dontSeeCheckboxIsChecked('#agree');
     * z.dontSeeCheckboxIsChecked('I agree to terms');
     * z.dontSeeCheckboxIsChecked('agree');
     * ```
     * @param {string} field  located by label|name|CSS|XPath|strict locator.
     */
    static async dontSeeCheckboxIsChecked(locator: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.dontSeeCheckboxIsChecked(locator[0], locator[1]);
            else
                await nonHeal.dontSeeCheckboxIsChecked(locator[0]);
        }
        else
            await nonHeal.dontSeeCheckboxIsChecked(locator);
    }



    /**
     * Scrolls element into viewport, waits for that element to become visible.
     * Then Checks that the given input field or textarea equals to given value.
     * 
     * ```js
     * z.seeInField('Username', 'davert');
     * z.seeInField({css: 'form textarea'},'Type your comment here');
     * z.seeInField('#searchform input','Search');
     * ```
     * @param {string} field located by label|name|CSS|XPath|strict locator.
     * @param {string} value value to check.
     */
    static async seeInField(locator: any, value: string) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.seeInField(locator[0], value, locator[1]);
            else
                await nonHeal.seeInField(locator[0], value);
        }
        else
            await nonHeal.seeInField(locator, value);
    }

    /**
    * Asserts that two values are not equal. If they are ,an assertion error is thrown. 
    * ```js
    * z.assertNotEqual(expected,actual)
    * ```
    * @param {any} expected 
    * @param {any} actual 
    */
    static async assertNotEqual(expected: any, actual: any, msg?: string) {
        await nonHeal.assertNotEqual(expected, actual, msg);
    }

    /**
    * Scrolls to element matched by locator. Extra shift can be set with offsetX and offsetY options.
    * ```js
    * z.scrollTo('footer');
    * z.scrollTo('#submit', 5, 5);
    
    * ```
    * @param {string} locator located by CSS|XPath|strict locator.
    * @param {number} offsetX (optional, 0 by default) X-axis offset
    * @param {number} offsetY (optional, 0 by default) Y-axis offset.
    *
    
    */
    static async scrollTo(locator: any, offsetX?: number, offsetY?: number) {
        if (Array.isArray(locator))
            await nonHeal.scrollTo(locator[0], offsetX, offsetY)
        else
            await nonHeal.scrollTo(locator, offsetX, offsetY)
    }

    /**
     * Scroll page to the top.
     * ```js
     * z.scrollPageToTop();
     
     */
    static async scrollPageToTop() {
        await nonHeal.scrollPageToTop();
    }

    /**
     * Scroll page to the bottom.
     * ```js
     * z.scrollPageToBottom();
     
     */
    static async scrollPageToBottom() {
        await nonHeal.scrollPageToBottom();
    }

    /**
    * Asserts that an element is visible a given number of times
    * ```js
    * await z..seeNumberOfVisibleElements('.buttons', 3);
    * ```
    * @param {string} locator CSS|XPath|strict locator.
    * @param {number} num number of elements..
    * 
    */
    static async seeNumberOfVisibleElements(locator: any, num: number) {
        if (Array.isArray(locator))
            await nonHeal.seeNumberOfVisibleElements(locator[0], num);
        else
            await nonHeal.seeNumberOfVisibleElements(locator, num);
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible.
     * Then Verifies that the specified checkbox is checked.
     * 
     * ```js
     * z.seeCheckboxIsChecked('Agree');
     * z.seeCheckboxIsChecked('#agree'); 
     * z.seeCheckboxIsChecked({css: '#signup_form input[type=checkbox]'});
     * ```
     * @param {string} field  located by label|name|CSS|XPath|strict locator.
     */
    static async seeCheckboxIsChecked(locator: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.seeCheckboxIsChecked(locator[0], locator[1]);
            else
                await nonHeal.seeCheckboxIsChecked(locator[0]);
        }
        else
            await nonHeal.seeCheckboxIsChecked(locator);
    }

    /**
    * Resize the current window to provided width and height. 
    * First parameter can be set to `maximize`.
    * ```js
    * z.resizeWindow(width, height)
    
    */
    static async resizeWindow(width: any, height?: any) {
        await nonHeal.resizeWindow(width, height);
    }

    /**
     * Types out the given text into an active field. 
     * To slow down typing use a second parameter, to set interval between key presses. 
     * Note: Should be used when fillField is not an option.
     * ```js
     * z.type('Type this out.');
     * 
     * // typing values with a 100ms interval
     * z.type('4141555311111111', 100);
     * // When passing in an array
     * z.type(['T', 'E', 'X', 'T'])  
     *       
     * @param {string} key|keys  key or array of keys to type.
     * @param {number} delay (optional) delay in ms between key presses
     */
    static async type(key: string | string[], delay?: number) {
        await nonHeal.type(key, delay);
    }

    /**
     * Performs a swipe Right inside an element. 
     * Can be slow or fast swipe.
     * ```js
     * z.swipeRight('#container');
     *       
     * @param {string} locator an element on which to perform swipe
     * @param {string}[speed='slow'] a speed to perform: `slow` or `fast`.
     */
    static async swipeRight(locator: any, speed?: string) {
        if (Array.isArray(locator))
            await nonHeal.swipeRight(locator[0], speed);
        else
            await nonHeal.swipeRight(locator, speed);
    }



    /**
     * Checks that title contains text.
     * ```js
     * z.seeInTitle('Home Page');
     * ```
     * @param {string} text value to check.
     */
    static async seeInTitle(text: any) {
        await nonHeal.seeInTitle(text);
    }

    /**
     * Scrolls element into viewport, waits for that element to become visible & enabled, then Force clicks an element
     
     * ```js
     * z.forceClick('#hiddenButton');
     * z.forceClick('Click me', '#hidden');
     * ```
    * @param {string} locator CSS|XPath|strict locator.
     * @param {string} context (optional, null by default) element to search in CSS|XPath|Strict locator.
     */
    static async forceClick(locator: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.forceClick(locator[0], locator[1]);
            else
                await nonHeal.forceClick(locator[0]);
        }
        else
            await nonHeal.forceClick(locator);
    }

    /**
     * Enters a directory In local filesystem.
     * Starts from a current directory
     * z.amInPath('output/downloads');
     * @param {string} openPath
     */
    static async amInPath(openPath: string) {
        await nonHeal.amInPath(openPath);
    }
    /**
     * Waits for file to be present in current directory.
     *
     * ```js
     * z.handleDownloads();
     * z.click('Download large File');
     * z.amInPath('output/downloads');
     * z.waitForFile('largeFilesName.txt', 10); // wait 10 seconds for file
     * ```
     * @param {string} name
     * @param {number} [sec] seconds to wait
     */
    static async waitForFile(name: string, sec?: number) {
        await nonHeal.waitForFile(name, sec);
    }

    /**
     * Checks that file with a name including given text exists in the current directory.
     *
     *```js
    * z.handleDownloads();
    * z.click('Download as PDF');
    * z.amInPath('output/downloads');
    * z.seeFileNameMatching('.pdf');
    * ```
    */
    static async seeFileNameMatching(text: string) {
        await nonHeal.seeFileNameMatching(text);
    }

    /**
     * Checks that file found by `seeFile` doesn't include text.
     * @param {string} text
     * @param {string} [encoding='utf8']
     */
    static async dontSeeInThisFile(text: string, encoding?: string) {
        await nonHeal.dontSeeInThisFile(text, encoding);
    }

    /**
    * Checks that all elements with given locator have given attributes.
    *
    * ```js
    * z.seeAttributesOnElements('//form', { method: "post"});
    * ```
    *
    * @param {CodeceptJS.LocatorOrString} locator located by CSS|XPath|strict locator.
    * @param {object} attributes attributes and their values to check.
    *
    */
    static async seeAttributesOnElements(locator: any, attributes: any) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.seeAttributesOnElements(locator[0], attributes, locator[1]);
            else
                await nonHeal.seeAttributesOnElements(locator[0], attributes);
        }
        else
            await nonHeal.seeAttributesOnElements(locator, attributes);
    }

    /**
     * Checks that a given Element is present in the DOM
     * Element is located by CSS or XPath.
     *
     * ```js
     * z.seeElementInDOM('#modal');
     * ```
     * @param {string} locator element located by CSS|XPath|strict locator.
     *
     */
    static async seeElementInDOM(locator: any) {
        if (Array.isArray(locator))
            await nonHeal.seeElementInDOM(locator[0]);
        else
            await nonHeal.seeElementInDOM(locator);
    }

    /**
    * Appends text into the field.
    * A field can be located by text, accessibility id, id.
    *
    * ```js
    * z.appendField('name', 'davert');
    * ```
    *
    * @param {string} field
    * @param {string} value
    */
    static async appendField(locator: any, value: string) {
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                await autoHeal.appendField(locator[0], value, locator[1]);
            else
                await nonHeal.appendField(locator[0], value);
        }
        else
            await nonHeal.appendField(locator, value);
    }

    /**
    * Waits for an element to become not attached to the DOM on a page (by default waits for 1sec).
    * Element can be located by CSS or XPath.
    *
    * ```js
    * z.waitForDetached('#popup');
    * ```
    *
    * @param {string} locator element located by CSS|XPath|strict locator.
    * @param {number} timeout (optional, `1` by default) time in seconds to wait
    */
    static async waitForDetached(locator: any, timeout?: number) {
        if (Array.isArray(locator))
            await nonHeal.waitForDetached(locator[0], timeout);
        else
            await nonHeal.waitForDetached(locator, timeout);
    }

    /**
     * Sends POST request to API.
     *
     * ```js
     * I.sendPostRequest('/api/users.json', { "email": "user@user.com" });
     * ```
     *
     * @param {*} url
     * @param {*} payload
     * @param {object} headers
     */
    static async sendPostRequest(url: any, payload?: any, headers?: any) {
        return await nonHeal.sendPostRequest(url, payload, headers)
    }

    /**
     * Send GET request to REST API
     *
     * ```js
     * I.sendGetRequest('/api/users.json');
     * ```
     *
     * @param {*} url
     * @param {object} headers
     */
    static async sendGetRequest(url: any, headers?: any) {
        return await nonHeal.sendGetRequest(url, headers);
    }

    /**
    * Opposite to `seeElementInDOM`. Checks that element is not on page.
    *
    * ```js
    * I.dontSeeElementInDOM('.nav'); // checks that element is not on page visible or not
    * ```
    *
    * @param {string} locator located by CSS|XPath|Strict locator.
    */
    static async dontSeeElementInDOM(locator: any) {
        if (Array.isArray(locator))
            await nonHeal.dontSeeElementInDOM(locator[0]);
        else
            await nonHeal.dontSeeElementInDOM(locator);
    }

    /**
     * Checks that text is equal to provided one.
     *
     * ```js
     * I.seeTextEquals('text', 'h1');
     * ```
     *
     * @param {string} text element value to check.
     * @param {string} optional [context]  element located by CSS|XPath|strict locator.
     */
    static async seeTextEquals(text: string, context?: any) {
        if (Array.isArray(context))
            await nonHeal.seeTextEquals(text, context[0]);
        else
            await nonHeal.seeTextEquals(text, context);
    }

    /**
     * Checks that title is equal to provided one.
     *
     * ```js
     * z.seeTitleEquals('Test title.');
     * ```
     *
     * @param {string} text value to check.
     */
    static async seeTitleEquals(text: string) {
        await nonHeal.seeTitleEquals(text);
    }

    /**
     * Clears a cookie by name,
     * if none provided clears all cookies.
     *
     * ```js
     * z.clearCookie();
     * z.clearCookie('test');
     * ```
     *
     * @param {string} (optional, `null` by default) cookie name
     * 
     */
    static async clearCookie(cookie?: string) {
        await nonHeal.clearCookie(cookie);
    }

    /**
      * Retrieves the innerHTML from an element located by CSS or XPath and returns it to test.
      * Resumes test execution, so **should be used inside async function with `await`** operator.
      * If more than one element is found - an array of HTMLs returned.
      *
      * ```js
      * let postHTML = await z.grabHTMLFrom('#post');
      * ```
      *
      * @param {CodeceptJS.LocatorOrString} element located by CSS|XPath|strict locator.
      * @returns {Promise<string>} HTML code for an element
      */
    // static async grabHTMLFrom(locator:string): Promise<string>
    // {
    //     let postHTML:string;

    //     return postHTML;
    // }

    /**
     * Retrieves all the innerHTML from elements located by CSS or XPath and returns it to test. 
     * Resumes test execution, so should be used inside async function with await operator.
     *
     * ```js
     * let postHTML = await z.grabHTMLFromAll('#post');
     * ```
     *
     * @param {CodeceptJS.LocatorOrString} element located by CSS|XPath|strict locator.
     * @returns {Promise<string>} HTML code for an element
     */
    static async grabHTMLFromAll(locator: any): Promise<any> {
        let postHTMLs: any[] = [];
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                postHTMLs = await autoHeal.grabHTMLFromAll(locator[0], locator[1]);
            else
                postHTMLs = await nonHeal.grabHTMLFromAll(locator[0]);
        }
        else
            postHTMLs = await nonHeal.grabHTMLFromAll(locator);
        return postHTMLs;
    }

    /**
    * Retrieves the cssProperty from an element located by CSS or XPath and returns it to test.
    * Resumes test execution, so **should be used inside async function with `await`** operator.
    * If more than one element is found - an array of HTMLs returned.
    *
    * ```js
    * const value = await I.grabCssPropertyFrom('h3', 'font-weight');
    * ```
    *
    * @param {CodeceptJS.LocatorOrString} element located by CSS|XPath|strict locator.
    *  @param {CodeceptJS.String} cssProperty CSS property name.
    * @returns {Promise<string>} CSS value
    */
    static async grabCssPropertyFrom(locator: any, cssProperty: string): Promise<string> {
        let postHTML: string = '';
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                postHTML = await autoHeal.grabCssPropertyFrom(locator[0], cssProperty, locator[1]);
            else
                postHTML = await nonHeal.grabCssPropertyFrom(locator[0], cssProperty);
        }
        else
            postHTML = await nonHeal.grabCssPropertyFrom(locator, cssProperty);
        return postHTML;
    }

    static async grabCssPropertyFromAll(locator: any, cssProperty: string): Promise<any> {
        let postHTML: any[] = [];
        if (Array.isArray(locator)) {
            if (process.env.AUTOHEAL == 'true' && (locator[0].startsWith("/") || locator[0].startsWith(".") || locator[0].startsWith("#") || locator[0].startsWith("{") || locator[0].startsWith("(")))
                postHTML = await autoHeal.grabCssPropertyFromAll(locator[0], cssProperty, locator[1]);
            else
                postHTML = await nonHeal.grabCssPropertyFromAll(locator[0], cssProperty);
        }
        else
            postHTML = await nonHeal.grabCssPropertyFromAll(locator, cssProperty);
        return postHTML;
    }

    static async generateLog(timestamp: number, error: any) {
        await nonHeal.generateLog(timestamp, error);
    }

    /**
     * Checks if element is visible and returns true of false accordingly.
     *
     * ```js
     * const value = await z.checkIfVisible("//input[@id='123']");
     * ```
     *
     * @param {string|object} locator element located by CSS|XPath|strict locator.
     * @returns {boolean} boolean
     */
    static async checkIfVisible(selector: any, ...options: any): Promise<any> {
        if (Array.isArray(selector))
            return await nonHeal.checkIfVisible(selector[0], ...options)
        else
            return await nonHeal.checkIfVisible(selector, ...options)
    }

    /**
     * Downloads the file from docker agent to code base agent.
     * We can also pass an expected file name to be set for the downloaded file.
     *
     * ```js
     * const fileName = await z.dockerFileDownload();
     * const fileName = await z.dockerFileDownload("test.pdf");
     * ```
     *
     * @returns {string} file path in the repo (./tempDownload/test.pdf)
     */
    static async dockerFileDownload(expectedFileName?: string) {
        await nonHeal.dockerFileDownload(expectedFileName);
    }

    /**
     * Starts the browser explicitly.
     *
     * ```js
     * await z.startBrowser();
     * ```
     */
    static async startBrowser() {
        try {
            console.log("========================>Start the Browser");
            await this.prototype.helpers["WebDriver"]._startBrowser();
        }
        catch (err) {
            console.log("========================>ERROR" + err);
            console.log("========================>Starting the Browser Failed...");
        }
    }

    /**
     * Stops the browser explicitly.
     *
     * ```js
     * await z.stopBrowser();
     * ```
     */
    static async stopBrowser() {
        try {
            console.log("========================>Stop the Browser");
            await this.prototype.helpers["WebDriver"]._stopBrowser();
        }
        catch (err) {
            console.log("========================>ERROR" + err);
            console.log("========================>Stopping the Browser Failed...");
        }
    }
}
